# ref: https://learn.microsoft.com/en-us/defender-xdr/malware-naming?view=o365-worldwide
from collections import defaultdict

PLATFORM = defaultdict(lambda: "")
PLATFORM['AndroidOS'] = "Android operating system"
PLATFORM['DOS'] = "MS-DOS platform"
PLATFORM['EPOC'] = "Psion devices"
PLATFORM['FreeBSD'] = "FreeBSD platform"
PLATFORM['iOS'] = "iPhone operating system"
PLATFORM['Linux'] = "Linux platform"
PLATFORM['macOS'] = "MAC 9.x platform or earlier"
PLATFORM['macOS_X'] = "macOS X or later"
PLATFORM['OS2'] = "OS2 platform"
PLATFORM['Palm'] = "Palm operating system"
PLATFORM['Solaris'] = "System V-based Unix platforms"
PLATFORM['SunOS'] = "Unix platforms 4.1.3 or lower"
PLATFORM['SymbOS'] = "Symbian operating system"
PLATFORM['Unix'] = "general Unix platforms"
PLATFORM['Win16'] = "Win16 (3.1) platform"
PLATFORM['Win2K'] = "Windows 2000 platform"
PLATFORM['Win32'] = "Windows 32-bit platform"
PLATFORM['Win64'] = "Windows 64-bit platform"
PLATFORM['Win95'] = "Windows 95, 98 and ME platforms"
PLATFORM['Win98'] = "Windows 98 platform only"
PLATFORM['WinCE'] = "Windows CE platform"
PLATFORM['WinNT'] = "WinNT"
PLATFORM['ABAP'] = "Advanced Business Application Programming scripts"
PLATFORM['ALisp'] = "ALisp scripts"
PLATFORM['AmiPro'] = "AmiPro script"
PLATFORM['ANSI'] = "American National Standards Institute scripts"
PLATFORM['AppleScript'] = "compiled Apple scripts"
PLATFORM['ASP'] = "Active Server Pages scripts"
PLATFORM['AutoIt'] = "AutoIT scripts"
PLATFORM['BAS'] = "Basic scripts"
PLATFORM['BAT'] = "Basic scripts"
PLATFORM['CorelScript'] = "Corelscript scripts"
PLATFORM['HTA'] = "HTML Application scripts"
PLATFORM['HTML'] = "HTML Application scripts"
PLATFORM['INF'] = "Install scripts"
PLATFORM['IRC'] = "mIRC/pIRC scripts"
PLATFORM['Java'] = "Java binaries (classes)"
PLATFORM['JS'] = "JavaScript scripts"
PLATFORM['LOGO'] = "LOGO scripts"
PLATFORM['MPB'] = "MapBasic scripts"
PLATFORM['MSH'] = "Monad shell scripts"
PLATFORM['MSIL'] = ".NET intermediate language scripts"
PLATFORM['Perl'] = "Perl scripts"
PLATFORM['PHP'] = "Hypertext Preprocessor scripts"
PLATFORM['Python'] = "Python scripts"
PLATFORM['SAP'] = "SAP platform scripts"
PLATFORM['SH'] = "Shell scripts"
PLATFORM['VBA'] = "Visual Basic for Applications scripts"
PLATFORM['VBS'] = "Visual Basic scripts"
PLATFORM['WinBAT'] = "Winbatch scripts"
PLATFORM['WinHlp'] = "Windows Help scripts"
PLATFORM['WinREG'] = "Windows registry scripts"
PLATFORM['A97M'] = "Access 97, 2000, XP, 2003, 2007, and 2010 macros"
PLATFORM['HE'] = "macro scripting"
PLATFORM['O97M'] = "Office 97, 2000, XP, 2003, 2007, and 2010 macros - those that affect Word, Excel, and PowerPoint"
PLATFORM['PP97M'] = "PowerPoint 97, 2000, XP, 2003, 2007, and 2010 macros"
PLATFORM['V5M'] = "Visio5 macros"
PLATFORM['W1M'] = "Word1Macro"
PLATFORM['W2M'] = "Word2Macro"
PLATFORM['W97M'] = "Word 97, 2000, XP, 2003, 2007, and 2010 macros"
PLATFORM['WM'] = "Word 95 macros"
PLATFORM['X97M'] = "Excel 97, 2000, XP, 2003, 2007, and 2010 macros"
PLATFORM['XF'] = "Excel formulas"
PLATFORM['XM'] = "Excel 95 macros"
PLATFORM['ASX'] = "XML metafile of Windows Media .asf files"
PLATFORM['HC'] = "HyperCard Apple scripts"
PLATFORM['MIME'] = "MIME packets"
PLATFORM['Netware'] = "Novell Netware files"
PLATFORM['QT'] = "Quicktime files"
PLATFORM['SB'] = "StarBasic (StarOffice XML) files"
PLATFORM['SWF'] = "Shockwave Flash files"
PLATFORM['TSQL'] = "MS SQL server files"
PLATFORM['XML'] = "XML files"


import os
import subprocess
from defender2yara.util.pe import parse_pe_resources

CAB_SIGNATURE = b'MSCF'


def expand_cab(cab_file_path):
    """
    Expands a CAB file using the Windows 'expand' command.

    Args:
        cab_file_path (str): The path to the CAB file.

    Raises:
        FileNotFoundError: If the CAB file does not exist.
        subprocess.CalledProcessError: If the expand command fails.
    """
    if not os.path.isfile(cab_file_path):
        raise FileNotFoundError(f"The CAB file '{cab_file_path}' does not exist.")
    
    output_dir = os.path.dirname(cab_file_path)

    command = ['expand', cab_file_path, '-F:*', output_dir]
    
    try:
        result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # print(result.stdout.decode("shift-jis"))
    except subprocess.CalledProcessError as e:
        print(f"Error expanding CAB file: {e.stderr.decode()}")
        raise e

def expand_mpam_fe(path):
    """
    Expand mpam-fe.exe using the Windows 'expand' command.

    Args:
        path (str): The path to the mpam-fe.exe file.
    """
    resources = parse_pe_resources(path)
    for resource in resources.values():
        for res in resource:
            if res['Data'][:4] == CAB_SIGNATURE:
                cab_data = res['Data']
    tmp_cab = os.path.join(os.path.dirname(path),"tmp.cab")
    with open(os.path.join(os.path.dirname(path),"tmp.cab"),"wb") as f:
        f.write(cab_data)
    expand_cab(tmp_cab)
    os.remove(tmp_cab)



from typing import List,Union,Tuple
import json

from defender2yara.defender.threat import Threat
from defender2yara.defender.constant import PLATFORM, SUFFIX
from defender2yara.defender.signature.hstr import HStrExtSig,HStrSig
from defender2yara.defender.subrule.hstr import HStrSubRule, HStrExtSubRule

from .strings import YaraString
from .condition import YaraCondition


class YaraRule:
    def __init__(self,threat:Threat,filesize_check,do_header_check:bool=False):
        self.threat = threat
        self.base_rule_name = self._gen_rule_name()
        self.meta = {}
        self._set_meta()
        self.rules:List[Tuple[Union[HStrSig,HStrExtSig],List[YaraString],YaraCondition]] = []
        
        signature:Union[HStrSig,HStrExtSig]
        for signature in threat.hstr_signatures:
            strings:List[YaraString] = []
            subrule:Union[HStrSubRule,HStrExtSubRule]
            for subrule in signature.subrules:
                rule_string = YaraString(subrule)
                if rule_string.string:
                    strings.append(rule_string)
            condition = YaraCondition(strings,signature,filesize_check,do_header_check)

            self.rules.append((signature,strings,condition))

    def _gen_rule_name(self) -> str:
        items = []
        if self.threat.threat_type:
            items.append(self.threat.threat_type)
        if self.threat.threat_platform:
            items.append(self.threat.threat_platform)
        if self.threat.threat_family:
            items.append(self.threat.threat_family)
        if self.threat.threat_variant:
            items.append(self.threat.threat_variant)
        if items:
            items.append(str(self.threat.threat_id))
            name:str = "_".join(items)
            if name[0].isdigit():
                name = "_" + name
            return name.replace("-","_")
        else:
            name = self.threat.threat_name\
                .replace("!","_")\
                .replace("/","_")\
                .replace("@","_")\
                .replace(":","_")\
                .replace(".","_")\
                .replace("-","_")\
                .replace("#","_")
            if name[0].isdigit():
                name = "_" + name
            return name

    def _set_meta(self):
        self.meta['author'] = "defender2yara"
        self.meta['detection_name'] = self.threat.threat_name
        self.meta['threat_id'] = str(self.threat.threat_id)
        if self.threat.threat_type:
            self.meta['type'] = self.threat.threat_type
        if self.threat.threat_platform:
            self.meta['platform'] = f"{self.threat.threat_platform}: {PLATFORM[self.threat.threat_platform]}"
        if self.threat.threat_family:
            self.meta['family'] = self.threat.threat_family
        if self.threat.severity:
            self.meta['severity'] = f"{self.threat.severity}"
        if self.threat.threat_suffixes:
            self.meta['info'] = []
            for suffix in self.threat.threat_suffixes:
                self.meta['info'].append(f"{suffix}: {SUFFIX[suffix]}")

    def generate_rules(self,add_comment=True):
        rule_idx = 0
        INDENT = " "*4
        for signature,strings,condition in self.rules:
            meta_section = []
            for key,value in self.meta.items():
                if isinstance(value,list):
                    for v in value:
                        meta_section.append(f"{key} = {json.dumps(v)}")
                else:
                    meta_section.append(f"{key} = {json.dumps(value)}")
            # add threshold info to meta section
            meta_section.append(f"signature_type = {json.dumps(signature.sig_type)}")
            meta_section.append(f"threshold = {json.dumps(str(condition.threshold))}")
            # add accuracy of the rule to meta section
            is_inaccurate = any([string.is_inaccurate for string in strings])
            meta_section.append(f"strings_accuracy = {json.dumps('High' if not is_inaccurate else 'Low')}")

            strings_section = []
            var_idx = 1
            for rule_string in strings:
                weight = rule_string.weight
                if add_comment and "hex" in rule_string.types:
                    line = f"${'x' if weight > 0 else 'n'}_{abs(weight)}_{var_idx} = {str(rule_string)} //weight: {weight}, accuracy: {'High' if not rule_string.is_inaccurate else 'Low'}"
                elif "hex" in rule_string.types:
                    line = f"${'x' if weight > 0 else 'n'}_{abs(weight)}_{var_idx} = {str(rule_string)}"
                elif add_comment:
                    line = f"${'x' if weight > 0 else 'n'}_{abs(weight)}_{var_idx} = {str(rule_string)} {' '.join(rule_string.types)} //weight: {weight}"
                else:
                    line = f"${'x' if weight > 0 else 'n'}_{abs(weight)}_{var_idx} = {str(rule_string)} {' '.join(rule_string.types)}"
                strings_section.append(line)
                var_idx += 1

            condition_section = []
            for statement in condition.statements:
                condition_section.append(statement)

            # construct YARA rule.
            rule = []
            rule.append(f"rule {self.base_rule_name}_{rule_idx}")
            rule.append("{")
            # write meta section
            rule.append(INDENT + "meta:")
            for line in meta_section:
                rule.append(INDENT*2 + line)
            # write strings section
            rule.append(INDENT + "strings:")
            for line in strings_section:
                rule.append(INDENT*2 + line)
            # write condition section
            rule.append(INDENT + "condition:")
            for i,statement in enumerate(condition_section):
                if isinstance(statement,str):
                    rule.append(INDENT*2 + "(" +statement + ")")
                
                if isinstance(statement,list):
                    rule.append(INDENT*2 + "(")
                    for j,sub in enumerate(statement):
                        rule.append(INDENT*3 + "(" + sub + ")")
                        if j != len(statement)-1:
                            rule[-1] += " or"
                    rule.append(INDENT*2 + ")")

                if i != len(condition_section)-1:
                    rule[-1] += " and"
            rule.append("}\n\n")
            rule_idx += 1
            yield "\n".join(rule)



from typing import Tuple
import platform
import httpx
import re
import time
from tqdm import tqdm
import os
import glob
import shutil

from defender2yara.util.cabarchive import expand_mpam_fe
from defender2yara.defender.vdm import Vdm

import logging

logger = logging.getLogger(__package__)

DOWNLOAD_URL = "https://go.microsoft.com/fwlink/?LinkID=121721&arch=x86"
URL_PATTERN_1 = r'https://definitionupdates\.microsoft\.com/download/DefinitionUpdates/versionedsignatures/[aA][mM]/[0-9.]+/[0-9.]+/x86/mpam-fe\.exe'
URL_PATTERN_2 = r'https://definitionupdates\.microsoft\.com/packages/content/mpam-fe\.exe\?packageType=Signatures&packageVersion=([0-9.]+)&arch=x86&engineVersion=([0-9.]+)'

UPDATE_CATALOG = "https://www.microsoft.com/en-us/wdsi/definitions/antimalware-definition-release-notes?requestVersion={signature_version}"

def check_cached_signature(signature_version,engine_version,cache_dir='cache') -> Tuple[bool,bool,bool]:
    has_base_signature = False
    has_delta_signature = False
    has_engine = False

    major_version = ".".join(signature_version.split('.')[0:2])
    minor_version = ".".join(signature_version.split('.')[2:4])

    base_vdm_dir = os.path.join(cache_dir,"vdm",major_version,"0.0")
    delta_vdm_dir = os.path.join(cache_dir,"vdm",major_version, minor_version)
    engine_dir = os.path.join(cache_dir,"engine",engine_version)

    av_base_path = os.path.join(base_vdm_dir,"mpavbase.vdm")
    as_base_path = os.path.join(base_vdm_dir,"mpasbase.vdm")
    av_delta_path = os.path.join(delta_vdm_dir,"mpavdlta.vdm")
    as_delta_path = os.path.join(delta_vdm_dir,"mpasdlta.vdm")
    mp_engine_path = os.path.join(engine_dir,"mpengine.dll")

    # @todo add exiftool check
    if os.path.exists(av_base_path) and os.path.exists(as_base_path):
        has_base_signature = True
    if os.path.exists(av_delta_path) and os.path.exists(as_delta_path):
        has_delta_signature = True
    if os.path.exists(mp_engine_path):
        has_engine = True
    return has_base_signature,has_delta_signature,has_engine


def download_file(url,cache_dir='cache',proxy=None) -> str:
    with httpx.stream("GET", url, proxy=proxy) as response:
        total = int(response.headers.get('content-length', 0))

        progress_bar = tqdm(
            total=total,
            unit='iB',
            bar_format='{l_bar}{bar:20}{r_bar}',
            colour='green',
            desc="Downloading latest signature",
            leave=False)

        output_path = os.path.join(cache_dir,"mpam-fe.exe")
        with open(output_path,'wb') as file:
            for chunk in response.iter_bytes():
                file.write(chunk)
                progress_bar.update(len(chunk))
        progress_bar.close()
        return output_path


def create_cache_dir(signature_version, engine_version, cache_path='cache')->None:
    os.makedirs(cache_path, exist_ok=True)
    
    major_version = ".".join(signature_version.split('.')[0:2])
    minor_version = ".".join(signature_version.split('.')[2:4])
    
    base_vdm_dir = os.path.join(cache_path,"vdm",major_version,"0.0")
    delta_vdm_dir = os.path.join(cache_path,"vdm",major_version, minor_version)
    engine_dir = os.path.join(cache_path,"engine",engine_version)

    os.makedirs(base_vdm_dir,exist_ok=True)
    os.makedirs(delta_vdm_dir,exist_ok=True)    
    os.makedirs(engine_dir,exist_ok=True)
    return


def move_files(signature_version,engine_version,source_dir,cache_path) -> None:
    major_version = ".".join(signature_version.split('.')[0:2])
    minor_version = ".".join(signature_version.split('.')[2:4])
    
    base_vdm_dir = os.path.join(cache_path,"vdm",major_version,"0.0")
    delta_vdm_dir = os.path.join(cache_path,"vdm",major_version, minor_version)
    engine_dir = os.path.join(cache_path,"engine",engine_version)

    # move base signature vmd
    shutil.move(os.path.join(source_dir,"mpasbase.vdm"),os.path.join(base_vdm_dir,"mpasbase.vdm"))
    shutil.move(os.path.join(source_dir,"mpavbase.vdm"),os.path.join(base_vdm_dir,"mpavbase.vdm"))

    # move delta signature vmd
    shutil.move(os.path.join(source_dir,"mpasdlta.vdm"),os.path.join(delta_vdm_dir,"mpasdlta.vdm"))
    shutil.move(os.path.join(source_dir,"mpavdlta.vdm"),os.path.join(delta_vdm_dir,"mpavdlta.vdm"))

    # move engine
    retry_count = 0
    while retry_count < 5:
        try:
            shutil.move(os.path.join(source_dir,"mpengine.dll"),os.path.join(engine_dir,"mpengine.dll"))
            break
        except PermissionError:
            time.sleep(1)
            retry_count += 1
    if retry_count == 5:
        logger.warning(f"Failed to move {os.path.join(source_dir,'mpengine.dll')}: PermissionError. (maybe due to antivirus scanning?)")
        #os.remove(os.path.join(source_dir,'mpengine.dll'))
    return


def get_latest_signature_vdm(proxy)->Tuple[str,str,str]:
    client = httpx.Client(proxy=proxy)
    res = client.head(DOWNLOAD_URL,follow_redirects=True)
    
    download_url = str(res.url)
    if re.match(URL_PATTERN_1,download_url):
        signature_version = download_url.split("/")[7]
        engine_version = download_url.split("/")[8]
        return download_url, signature_version, engine_version
    elif re.match(URL_PATTERN_2,download_url):
        m = re.match(URL_PATTERN_2,download_url)
        signature_version = m.groups()[0]
        engine_version = m.groups()[1]
        return download_url, signature_version, engine_version
    return None,None,None


def parse_mpam_exe(full_engine_path:str,cache_path:str,rm_mpam:bool) -> Tuple[str,str]:
    if not os.path.exists(full_engine_path):
        raise FileNotFoundError(f"mpam-fe.exe file not found: {full_engine_path}")
    if platform.system() == 'Windows':
        # extract cabarchive with windows expand command
        expand_mpam_fe(full_engine_path)
        source_dir = os.path.dirname(full_engine_path)
    else:
        # extract cabarchive with libarchive
        import libarchive
        libarchive.extract_file(full_engine_path)
        source_dir = os.path.curdir

    vdm_path = os.path.join(source_dir,"mpavdlta.vdm")
    engine_path = os.path.join(source_dir,"mpengine.dll")

    _, signature_version = Vdm.get_meta_info(vdm_path)
    _, engine_version = Vdm.get_meta_info(engine_path)

    # create cache dir
    create_cache_dir(signature_version,engine_version,cache_path)

    # move files
    move_files(signature_version,engine_version,source_dir,cache_path)

    # clean up
    files_to_remove = glob.glob("M?SigStub.exe",root_dir=source_dir)
    for file_path in files_to_remove:
        os.remove(os.path.join(source_dir,file_path))

    if rm_mpam:
        os.remove(full_engine_path)
    
    return signature_version,engine_version


def download_latest_signature(cache_path='cache',proxy=None) -> Tuple[str,str,bool]:
    use_cache = False
    os.makedirs(cache_path, exist_ok=True)
    download_url, signature_version, engine_version = get_latest_signature_vdm(proxy=proxy)
    if not download_url:
        raise ConnectionError(f"Failed to fetch Signature download URL:{DOWNLOAD_URL}")

    # check db files
    has_base_signature,\
    has_delta_signature,\
    has_engine = check_cached_signature(signature_version, engine_version, cache_path)

    if not (has_base_signature and has_delta_signature and has_engine):
        dl_file_path = download_file(download_url,proxy=proxy)
        if not os.path.exists(dl_file_path):
            raise FileNotFoundError(f"Download file not found: {dl_file_path}")
        signature_version,engine_version = parse_mpam_exe(dl_file_path,cache_path,rm_mpam=True)
    else:
        use_cache = True

    return signature_version,engine_version,use_cache



from typing import Tuple
import os
import pefile

import logging

logger = logging.getLogger(__package__)

def format_version(version: int) -> str:
    """
    Format a PE version into a string.

    Args:
        version (int): The PE version to format.

    Returns:
        str: The formatted version string.
    """
    major = version >> 16
    minor = version & 0xFFFF
    return f"{major}.{minor}"

def parse_pe_meta_info(pe_file_path: str) -> Tuple[str,str]:
    """
    Parse the metadata from a PE file.

    Args:
        pe_file_path (str): The path to the PE file.

    Returns:
        Tuple[Optional[str], Optional[str]]: A tuple containing the original filename and product version.
        If an error occurs, returns ("", "").
    """
    try:
        pe = pefile.PE(pe_file_path)
        # get product version info
        if hasattr(pe, 'VS_FIXEDFILEINFO'):
            version_info = pe.VS_FIXEDFILEINFO[0]
            version_data = {
                "ProductVersionMS": version_info.ProductVersionMS,
                "ProductVersionLS": version_info.ProductVersionLS,
            }
            product_version = \
                format_version(version_data["ProductVersionMS"]) + \
                "." + \
                format_version(version_data["ProductVersionLS"])
        else:
            product_version = None
        
        # get original filename
        original_filename = None
        for file_info in pe.FileInfo:
            for entry in file_info:
                if hasattr(entry, "StringTable"):
                    for st_entry in entry.StringTable:
                        for key, entry in list(st_entry.entries.items()):
                            #offsets = st_entry.entries_offsets[key]
                            lengths = st_entry.entries_lengths[key]
                            st_data: str
                            if len(entry) > lengths[1]:
                                st_data = entry.decode("utf-8")
                            else:
                                st_data = entry.decode("utf-8")
                            if st_data.endswith((".exe", ".dll", ".vdm")):
                                original_filename = st_data
        return original_filename, product_version
    except Exception as e:
        logger.warning(f"Failed to parse metadata from PE file. Error: {e}")
        raise e

def parse_pe_resources(pe_file_path):
    """
    Parses the resources section of a PE file using the pefile library.

    Args:
        pe_file_path (str): The path to the PE file.

    Returns:
        dict: A dictionary containing the parsed resources.

    Raises:
        FileNotFoundError: If the PE file does not exist.
        pefile.PEFormatError: If the file is not a valid PE file.
    """
    resources = {}
    
    if not os.path.isfile(pe_file_path):
        raise FileNotFoundError(f"The PE file '{pe_file_path}' does not exist.")
    
    try:
        pe = pefile.PE(pe_file_path)
    except pefile.PEFormatError as e:
        raise pefile.PEFormatError(f"Error parsing PE file: {e}")
    
    if hasattr(pe, 'DIRECTORY_ENTRY_RESOURCE'):
        for resource_type in pe.DIRECTORY_ENTRY_RESOURCE.entries:
            type_name = pefile.RESOURCE_TYPE.get(resource_type.struct.Id, str(resource_type.struct.Id))
            resources[type_name] = []
            
            if hasattr(resource_type, 'directory'):
                for resource_id in resource_type.directory.entries:
                    if hasattr(resource_id, 'directory'):
                        for resource_lang in resource_id.directory.entries:
                            data_rva = resource_lang.data.struct.OffsetToData
                            size = resource_lang.data.struct.Size
                            data = pe.get_memory_mapped_image()[data_rva:data_rva+size]
                            resources[type_name].append({
                                'ResourceId': resource_id.struct.Id,
                                'Language': resource_lang.struct.Id,
                                'Data': data
                            })
    else:
        raise ValueError("No resources found in the PE file.")
    return resources


from typing import Tuple, List, Union
import json
import re
from defender2yara.defender.subrule.hstr import HStrSubRule,HStrExtSubRule
from defender2yara.util import is_printable_ascii, is_printable_utf16_le_ascii

class YaraString:
    def __init__(self,subrule:Union[HStrSubRule,HStrExtSubRule]):
        self.string,\
        self.types,\
        self.is_inaccurate = self.parse_subrule(subrule.rule_bytes)
        self.weight = subrule.rule_weight

    def __repr__(self) -> str:
        if "ascii" in self.types or "wide" in self.types:
            return json.dumps(self.string)
        if "hex" in self.types:
            return f"{{{self.string}}} "

    @staticmethod
    def parse_subrule(data:bytes):
        """
        Converts rule bytes to a YARA string format.

        Args:
        data (bytes): The rule bytes to be converted.

        Returns:
            Tuple[str, List[str], bool]:
                - str: The YARA string.
                - List[str]: The types of YARA strings identified (e.g., 'ascii', 'wide', 'hex').
                - bool: Flag indicating if the YARA string conversion is inaccurate.
        """
        string = ""
        string_types = []
        is_inaccurate = False

        if is_printable_ascii(data):
            string = data.decode('utf-8')
            string_types.append('ascii')
        elif is_printable_utf16_le_ascii(data):
            string = data.decode('utf-16-le')
            string_types.append('wide')
        elif len(data) > 2 and data[-1] == 0x00 and data[-2] == 0x90:
            string, is_inaccurate = YaraString.parse_ext_subrule(data)
            string_types.append('hex')
            is_inaccurate = True
        else:
            string = ' '.join(f'{byte:02x}' for byte in data)
            string_types.append('hex')

        return string, string_types, is_inaccurate

    @staticmethod
    def parse_ext_subrule(data:bytes,is_root:bool=True)->Tuple[str,bool]:
        """
        Converts HStrExt data to a YARA hex string.

        Args:
            data (bytes): The rule bytes stored in HStrExtSubRule.
            is_root (bool): Boolean to control recursive call.

        Returns:
            Tuple[str, bool]:
                - str: The formatted string in YARA hex style.
                - bool: Whether an unimplemented regex-like pattern was encountered.

        Raises:
            ValueError: If the data is too short or contains invalid suffix data.
        """
        is_inaccurate = False
        WILDCARD_MARKER = 0x90

        if len(data) < 3:
            if is_root:
                raise ValueError("Too short data.",data)
            else:
                return  " ".join(f"{c:02x}" for c in data), is_inaccurate
        
        if is_root and (data[-1] != 0x00 or data[-2] != 0x90):
            raise ValueError(f"Invalid suffix data: '{data[-2]:02x} {data[-1]:02x}'")

        p = 0
        hex_string = []
        while p < len(data)-2:
            if data[p] == WILDCARD_MARKER:
                if data[p+1] == 0x01: # static length wildcard
                    wildcard_size = data[p+2]
                    hex_string.extend(["??" for _ in range(wildcard_size)])
                    p += 3

                elif data[p+1] == 0x02: # dynamic length wildcard
                    if is_root:
                        upper_size = data[p+2]
                        hex_string.append(f"[0-{upper_size}]")
                    else:
                        # Nested dynamic length is not supported
                        # e.g.: 
                        #    { aa (de ad be ef|[0-3]) bb } <- [0-3] is Invalid
                        # here we convert hex pattern into following syntax
                        #    { aa (de ad be ef|(?? | ?? ?? | ?? ?? ??) bb }
                        # however this syntax is too complex for yara hex pattern
                        is_inaccurate = True
                        upper_size = data[p+2]
                        if upper_size > 0:
                            _tmp = "("
                            for i in range(1,upper_size):
                                _tmp += " ".join(["??" for _ in range(i)])
                                _tmp += "|"
                            if len(_tmp) > 2:
                                _tmp = _tmp[:-1]+")"
                                hex_string.append(_tmp)
                    p += 3

                elif data[p+1] == 0x03: # two byte sequence patter
                    first_seq_size = data[p+2]
                    second_seq_size = data[p+3]
                    first_bytes = data[p+4:p+4+first_seq_size]
                    second_bytes = data[p+4+first_seq_size:p+4+first_seq_size+second_seq_size]

                    first_bytes_yara_str, _is_inaccurate = YaraString.parse_ext_subrule(first_bytes,False)
                    is_inaccurate |= _is_inaccurate

                    second_bytes_yara_str, _is_inaccurate = YaraString.parse_ext_subrule(second_bytes,False)
                    is_inaccurate |= _is_inaccurate

                    if first_bytes_yara_str and second_bytes_yara_str:
                        hex_string.append(f"({first_bytes_yara_str}|{second_bytes_yara_str})")
                    elif first_bytes_yara_str:
                        hex_string.append(f"{first_bytes_yara_str}")
                    elif second_bytes_yara_str:
                        hex_string.append(f"{second_bytes_yara_str}")

                    p += 4 + first_seq_size + second_seq_size

                elif data[p+1] == 0x04: # case sensitive regex-like style
                    iterate_count = data[p+2]
                    pattern_size = data[p+3]
                    pattern_bytes = data[p+4:p+4+pattern_size]
                    pattern_yara_str = "|".join(f"{c:02x}" for c in pattern_bytes)
                    # reduce regex complexity
                    if iterate_count > 2:
                        hex_string.extend(["??" for _ in range(iterate_count-2)])
                        hex_string.extend([f"({pattern_yara_str})" for _ in range(2)])
                    else:
                        hex_string.extend([f"({pattern_yara_str})" for _ in range(iterate_count)])
                    p += 4 + pattern_size

                elif data[p+1] == 0x05: # case insensitive regex-like style
                    upper_size = data[p+2]
                    pattern_size = data[p+3]
                    pattern_bytes:bytes = data[p+4:p+4+pattern_size]
                    pattern_yara_str = "|".join(f"{c:02x}" for c in pattern_bytes)
                    # hex_string.append(f"({pattern_yara_str})") 
                    ## Since Yara 4.4 or earlier does not have the capability to express "iterate hex pattern"
                    hex_string.append(f"[0-{upper_size*pattern_size}]")
                    p += 4 + pattern_size

                elif data[p+1] <= 0x20:
                    is_inaccurate = True
                    p += 2

                elif data[p+1] == 0x90: # maybe magic byte escape
                    hex_string.append(f"{data[p+1]:02x}")
                    p += 2

                else:
                    is_inaccurate = True
                    hex_string.append(f"{data[p+1]:02x}")
                    hex_string.append(f"{data[p+2]:02x}")
                    p += 2
            else:
                hex_string.append(f"{data[p]:02x}")
                p += 1

        result = " ".join(hex_string)

        # fix unsupported regex-like syntax #@todo make regex-like syntax check
        # e.g.) {57 3e 00 00 02 00 81 [0-3]} <- hex strings ending with "[0-3]" is invalid.
        result = re.sub(r' \[0-\d+\]$', '', result) 
        # e.g.) {[0-3] 57 3e 00 00 02 00 81} <- hex strings starting with "[0-3]" is invalid.
        result = re.sub(r'^\[0-\d+\]', '', result) 

        return result, is_inaccurate



from typing import List,Dict,Tuple
import os
import sys
import shutil
import yara
from collections import defaultdict

from defender2yara.defender.threat import Threat
from defender2yara.yara.rule import YaraRule
from defender2yara.defender.vdm import Vdm
from defender2yara.defender.download import get_latest_signature_vdm, download_latest_signature, parse_mpam_exe

from tqdm import tqdm
import logging

logger = logging.getLogger(__package__)

def clean_up_dir(path:str):
    if not os.path.exists(path):
        raise ValueError(f"Path does not exists: {path}")
    
    for root, dirs, files in os.walk(path):
        for file in files:
            file_path = os.path.join(root, file)
            try:
                os.remove(file_path)
            except Exception as e:
                logger.error(f"Failed to delete: {file_path}. Error: {e}")
        
        for dir in dirs:
            dir_path = os.path.join(root, dir)
            try:
                shutil.rmtree(dir_path)
            except Exception as e:
                logger.error(f"Failed to delete directory: {dir_path}. Error: {e}")


def write_rules_to_single_file(path,filename,results:Dict[Threat,List[str]]):
    output = []
    for rules in results.values():
        output.extend(rules)
    # write result
    output_file = os.path.join(path,f"{filename}.yara")        
    with open(output_file,'w') as f:
        f.writelines(output)
        logger.info(f"Write YARA rules to {output_file}")


def write_rules_by_family(path,results:Dict[Threat,List[str]]):
    threat:Threat
    for threat,rules in results.items():
        output_dir = os.path.join(path,threat.threat_platform,threat.threat_type)
        if not threat.threat_family:
            output_file = os.path.join(output_dir,"misc.yara")
        else:
            output_file = os.path.join(output_dir,threat.threat_family+".yara")

        os.makedirs(output_dir, exist_ok=True)
        with open(output_file,'a') as f:
            f.writelines(rules)
    logger.info(f"Write YARA rules to {path}")


def covert_vdm_to_yara(vdm:Vdm,filesize_check:str,header_check:bool=False) -> Tuple[Dict[Threat,List[str]],int]:
    logger.info(f"Parsing signature database...")
    results:Dict[Threat,List[str]] = defaultdict(list)
    rule_count = 0
    # convert to yara rule
    threats = vdm.get_threats()

    progress_bar = tqdm(
            total=len(threats),
            unit='threat',
            bar_format='{l_bar}{bar:20}{r_bar}',
            colour='green',
            desc="Converting signatures",
            leave=False)

    for threat in threats:
        yara_rules = YaraRule(threat,filesize_check=filesize_check,do_header_check=header_check)
        if not yara_rules:
            continue
        for yara_rule in yara_rules.generate_rules():
            try:
                yara.compile(source=yara_rule)
            except yara.SyntaxError as e:
                logger.warn(f"Failed to convert {threat.threat_name}: {str(e)}")
                logger.debug("\n"+yara_rule)
                continue
            results[threat].append(yara_rule)
            rule_count += 1
        progress_bar.update(1)

    progress_bar.close()

    return results,rule_count


def main(args):
    cache_dir = args.cache
    signature_version:str = ""
    engine_version:str = ""

    if args.latest_signature_version:
        url, signature_version, engine_version = get_latest_signature_vdm(proxy=args.proxy)
        print(f"{signature_version}")
        sys.exit(0)

    if (args.download or not args.base) and not args.mpam:
        logger.info("Downloading latest signature database.")
        signature_version, engine_version, use_cache = download_latest_signature(cache_dir,proxy=args.proxy)
        logger.info(f"Complete (use_cache:{use_cache})")
        logger.info(f"Latest Signature Version:{signature_version}")
        logger.info(f"Latest Engine Version   :{engine_version}")

    if args.download:
        sys.exit(0)

    if args.mpam:
        signature_version, engine_version = parse_mpam_exe(args.mpam,cache_path=cache_dir,rm_mpam=False)
        logger.info(f"Loaded {args.mpam}")
        logger.info(f"Latest Signature Version:{signature_version}")
        logger.info(f"Latest Engine Version   :{engine_version}")

    base_file:str = ""
    delta_file:str = ""
    results:Dict[Threat,List[str]]

    output_path = os.path.join(args.output,signature_version)
    os.makedirs(output_path, exist_ok=True)
    logger.info(f"Clean up output directory: {output_path}")
    clean_up_dir(output_path)

    if args.base: # use manually specified vdm files.
        logger.info(f"Loading base signature file: {args.base}")
        vdm = Vdm(args.base)
        if args.delta:
            logger.info(f"Applying delta patch: {args.delta}")
            vdm.apply_delta_vdm(args.delta)

        logger.info(f"Target signature version: {vdm.version}")
        logger.info(f"Target signature type   : {vdm.vdm_type}")

        results,rule_counts = covert_vdm_to_yara(vdm,args.filesize_check,args.header_check)
        logger.info(f"Convert {rule_counts} signatures.")

        if args.single_file:
            write_rules_to_single_file(output_path,vdm.vdm_type,results)
        else:
            write_rules_by_family(output_path,results)
    else: # use vdm files parsed from mpam-fe.exe
        major_version = ".".join(signature_version.split(".")[0:2])
        minor_version = ".".join(signature_version.split(".")[2:4])
        vdm_base_path = os.path.join(cache_dir,"vdm",major_version,'0.0')
        vdm_delta_path = os.path.join(cache_dir,"vdm",major_version,minor_version)
        
        for name in ["mpav","mpas"]:
            base_file = os.path.join(vdm_base_path,name+"base.vdm")
            delta_file = os.path.join(vdm_delta_path,name+"dlta.vdm")
        
            logger.info(f"Loading base signature file: {base_file}")
            vdm = Vdm(base_file)
            if os.path.exists(delta_file):
                logger.info(f"Applying delta patch: {delta_file}")
                vdm.apply_delta_vdm(delta_file)

            logger.info(f"Target signature version: {vdm.version}")
            logger.info(f"Target signature type   : {vdm.vdm_type}")

            results,rule_counts = covert_vdm_to_yara(vdm,args.filesize_check,args.header_check)
            logger.info(f"Convert {rule_counts} signatures")

            if args.single_file:
                write_rules_to_single_file(output_path,vdm.vdm_type,results)
            else:
                write_rules_by_family(output_path,results)

    logger.info("Complete")
    sys.exit(0)



import struct

from .base import BaseSig
from defender2yara.defender.constant import SIG_TYPES, THREAT_NAME_PREFIX


class ThreatBeginSig(BaseSig):
    def __init__(self,data:bytes,ptr:int=0) -> None:
        # parse common header
        self.parse_common_header(data,ptr)
        assert self.sig_type == SIG_TYPES[0x5c]

        # parse threat info
        self.threat_id = struct.unpack('<I', self.sig_data[0:4])[0]
        self.unknown_bytes1 = self.sig_data[4:8]
        self.category_id = struct.unpack('<H', self.sig_data[8:10])[0]
        self.size_threat_name = self.sig_data[10]
        self.type_id = self.sig_data[11] # not concrete
        self.raw_threat_name = self.sig_data[12:12+self.size_threat_name]
        
        # parse severity_id
        try:
            self.unknown_bytes2 = self.sig_data[12+self.size_threat_name:16+self.size_threat_name]
            self.severity_id = self.sig_data[16+self.size_threat_name]
            self.unknown_bytes3 = self.sig_data[17+self.size_threat_name:17+self.size_threat_name+4]
        except IndexError:
            self.unknown_bytes2 = None
            self.severity_id = None
            self.unknown_bytes3 = None

        # resolve threat name
        if self.raw_threat_name[0] > 128:
            prefix_id = struct.unpack('<H', self.raw_threat_name[0:2])[0]
            self.threat_name = THREAT_NAME_PREFIX[prefix_id] + self.raw_threat_name[2:].decode('utf-8')
        else:
            self.threat_name = self.raw_threat_name.decode('utf-8')
        


class ThreatEndSig(BaseSig):
    def __init__(self,data:bytes,ptr:int=0) -> None:
        # parse common header
        self.parse_common_header(data,ptr)
        assert self.sig_type == SIG_TYPES[0x5D]

        self.threat_id = struct.unpack('<I', self.sig_data[0:4])[0]



import struct
# from defender2yara.utils import hexdump


class HStrSubRule:
    def __init__(self,data:bytes,ptr:int) -> None:
        if len(data[ptr:]) < 3:
            self.rule_weight = 0
            return

        self.rule_weight = struct.unpack('<h',data[ptr+0:ptr+2])[0] # rule weight contains negative numbers

        if self.rule_weight == 0:
            return
        self.rule_size = data[ptr+2]
        self.rule_bytes = data[ptr+3:ptr+3+self.rule_size]

        # Convert rule_bytes to YARA strings.
        # self.yara_str, self.yara_str_type, self.yara_str_inaccurate = rule_bytes_to_yara_string(self.rule_bytes)


class HStrExtSubRule:
    def __init__(self,data:bytes,ptr:int) -> None:
        if len(data[ptr:]) < 4:
            self.rule_weight = 0
            return 

        self.rule_weight = struct.unpack('<h',data[ptr+0:ptr+2])[0] # rule weight contains negative numbers

        if self.rule_weight == 0:
            return

        self.rule_size = data[ptr+2]
        self.unknown_u8_1 = data[ptr+3]

        if self.unknown_u8_1 in [0x80,0x81,0x88]: # unknown flag.
            self.rule_bytes = data[ptr+5:ptr+5+self.rule_size]
            self.rule_size += 1
        elif self.unknown_u8_1 == 0x90: # unknown flag.
            self.rule_size += 1
            self.rule_bytes = data[ptr+5:ptr+5+self.rule_size]
        else:
            self.rule_bytes = data[ptr+4:ptr+4+self.rule_size]

        # Convert rule_bytes to YARA strings.
        # self.yara_str, self.yara_str_type, self.yara_str_inaccurate = rule_bytes_to_yara_string(self.rule_bytes)


from typing import List,Dict,Union

from collections import Counter
from defender2yara.defender.signature.hstr import HStrSig,HStrExtSig
from defender2yara.util.utils import all_elements_equal

from .strings import YaraString

class YaraCondition:
    def __init__(self,strings:List[YaraString],signature:Union[HStrSig,HStrExtSig],filesize_check:str,do_header_check:bool=False):
        self.threshold = signature.threshold
        self.strings = strings
        self.signature = signature
        self.do_header_check = do_header_check
        self.filesize_check = filesize_check
        self.statements:List[Union[str,List[str]]] = [] # all state is connected with "and"
        self.compose_logical_expression()

    def compose_logical_expression(self):
        positive_weight_list = [string.weight for string in self.strings if string.weight > 0]
        positive_weight_sum = sum(positive_weight_list)

        # add filesize check to improve the scan performance
        if self.filesize_check:
            self.statements.append(f"filesize < {self.filesize_check}")

        # add file header check to improve the scan performance
        if self.do_header_check:
            if "PEHSTR" in self.signature.sig_type or "DOSHSTR" in self.signature.sig_type:
                self.statements.append("uint16(0) == 0x5a4d")
            elif "MACHOHSTR" in self.signature.sig_type:
                self.statements.append(
                    "(uint32(0) == 0xfeedfacf) or "
                    "(uint32(0) == 0xcffaedfe) or "
                    "(uint32(0) == 0xfeedface) or "
                    "(uint32(0) == 0xcefaedfe)"
                ) 
            elif "ELFHSTR" in self.signature.sig_type:
                self.statements.append("uint32(0) == 0x464c457f")

        # to simplify the problem, formulate the condition to exclude any of elements with negative weights.
        if min([string.weight for string in self.strings]) < 0:
            self.statements.append("not (any of ($n*))")

        # solve simple cases
        if self.threshold >= positive_weight_sum:
            self.statements.append("all of ($x*)")
            return
        elif all_elements_equal(positive_weight_list) and self.threshold != positive_weight_sum:
            self.statements.append(f"{(self.threshold + positive_weight_list[0] - 1)//positive_weight_list[0]} of ($x*)")
            return

        # solve complex cases
        statements_or:List[str] = []
        positive_weight_map = dict(sorted(Counter(positive_weight_list).items(), key=lambda x: x[0], reverse=True))
        subsets = self.generate_subsets(positive_weight_map,self.threshold)
        for subset in subsets:
            _tmp = []
            weight_count = Counter(subset)
            for w,cnt in weight_count.items():
                _tmp.append(f"{cnt} of ($x_{w}_*)")
            condition = "(" + " and ".join(_tmp) + ")"
            statements_or.append(condition)
        statements_or.append("all of ($x*)") # to avoid yara compile error
        self.statements.append(statements_or)

    
    @staticmethod
    def generate_subsets(weight_map: Dict[int, int], threshold: int) -> List[List[int]]:
        """
        Generate all subsets of weights that meet or exceed the given threshold, 
        without including supersets of already valid subsets.

        This function uses a backtracking algorithm to generate all possible subsets
        of weights from the given weight map, where the sum of weights in each subset
        is greater than or equal to the specified threshold. It includes all minimal
        subsets that meet the threshold, but excludes any superset of an already 
        included subset.

        Args:
            weight_map (Dict[int, int]): A dictionary mapping weights to their counts.
            threshold (int): The minimum sum of weights required for a valid subset.

        Returns:
            List[List[int]]: A list of all valid subsets, where each subset is 
                            represented as a list of weights. If a subset meets the
                            threshold, no superset of it will be included in the result.

        Example:
            >>> weight_map = {5: 2, 3: 1, 2: 2, 1: 1}
            >>> threshold = 7
            >>> generate_subsets(weight_map, threshold)
            [[5, 5], [5, 3], [5, 2, 2]]
            # Note: [5, 5, 1] is not included as [5, 5] already meets the threshold
        """
        # Get lists of keys (weights) and values (counts) from the dictionary
        weight_values = list(weight_map.keys())
        weight_counts = list(weight_map.values())

        def backtrack(index: int, current_subset: List[int], subsets: List[List[int]]):
            """
            Helper function to generate all subsets using backtracking with pruning.

            This function implements a depth-first search with backtracking to generate
            all valid subsets. It uses two key pruning strategies to optimize the search:

            1. Early termination: If the current subset sum meets or exceeds the threshold,
            we add it to the result (if valid) and stop exploring further additions.

            2. Subset pruning: If we find a valid subset, we skip adding any elements
            that would create a superset of an already valid subset. This prevents
            generating redundant supersets.

            Args:
                index (int): Current index in the weight_values list.
                current_subset (List[int]): The current subset being built.
                subsets (List[List[int]]): List to store all valid subsets.

            Note:
                The function modifies the 'subsets' list in-place to collect valid subsets.
                It uses recursion to explore all possible combinations of weights.
            """
            # If the sum of the current subset meets or exceeds the threshold
            if sum(current_subset) >= threshold:
                # Don't add if a subset that sums to the threshold has already been found
                if sum(current_subset) - current_subset[-1] >= threshold:
                    return
                subsets.append(current_subset.copy())
                return

            # If we've reached the end of the weight list, return
            if index == len(weight_values):
                return

            # Case 1: Don't include the current weight
            backtrack(index + 1, current_subset, subsets)

            # Case 2: Include the current weight (up to its count)
            for _ in range(weight_counts[index]):
                current_subset.append(weight_values[index])
                backtrack(index + 1, current_subset, subsets)

            # Restore the original state (backtrack)
            for _ in range(weight_counts[index]):
                current_subset.pop()

        # List to store all valid subsets
        all_subsets = []
        backtrack(0, [], all_subsets)
        return all_subsets


from typing import List, Tuple

import os
import struct
import zlib
from collections import defaultdict
import exiftool

from defender2yara.defender.constant import SIG_TYPES
from defender2yara.defender.signature import *
from defender2yara.defender.threat import Threat

from defender2yara.util.utils import hexdump
from defender2yara.util.pe import parse_pe_meta_info

import logging
from tqdm import tqdm

logger = logging.getLogger(__package__)

class Vdm:
    def __init__(self,base_path:str):
        self.base_path = base_path
        self.delta_path = ""

        self.signatures:List[BaseSig] = []
        self.threats:List[Threat] = []

        self.delta_filename:str = ""
        self.version:str = ""
        self.vdm_type:str = ""

        filename, self.version = self.get_meta_info(base_path)

        if filename not in ['mpavbase.vdm','mpasbase.vdm']:
            raise ValueError("Invalid vdm file. Please use mpa(s|v)base.vdm file.")

        if filename.startswith("mpav"):
            self.vdm_type = "anti-virus"
        elif filename.startswith("mpas"):
            self.vdm_type = "anti-spyware"

        self.major_version = ".".join(self.version.split(".")[0:2])
        self.minor_version = ".".join(self.version.split(".")[2:4])

        self.raw_data:bytes = self.extract_vdm_data(base_path)


    def apply_delta_vdm(self,vdm_path):
        delta_filename, version = self.get_meta_info(vdm_path)
        # Validation
        if delta_filename not in ['mpavdlta.vdm','mpasdlta.vdm']:
            raise ValueError(f"Invalid vdm file. Please use mpa(s|v)dlta.vdm file:{delta_filename}")
        
        vdm_type:str = ""
        if delta_filename.startswith("mpav"):
            vdm_type = "anti-virus"
        elif delta_filename.startswith("mpas"):
            vdm_type = "anti-spyware"

        if vdm_type != self.vdm_type:
            raise ValueError(f"Vdm file type miss match: Base->{self.vdm_type}, Delta->{vdm_type}")

        major_version = ".".join(version.split(".")[0:2])
        minor_version = ".".join(version.split(".")[2:4])
        
        if major_version != self.major_version:
            raise ValueError(f"The major version of vdm miss matched: Base->{self.major_version}, Delta:{major_version}")

        self.delta_path = vdm_path
        self.version = ".".join([self.major_version,minor_version])
        self.minor_version = minor_version
    
        delta_raw_data = self.extract_vdm_data(vdm_path)
        delta_signatures = self.parse_database(delta_raw_data)

        if len(delta_signatures) != 2:
            raise ValueError(f"Unexpected signature count for delta vdm: {len(delta_signatures)}")

        delta_blob = delta_signatures[1].sig_data
        self.raw_data = self.apply_delta_patch(delta_blob,self.raw_data)
    
    def get_signatures(self) -> List[BaseSig]:
        self.signatures = self.parse_database(self.raw_data)
        return self.signatures

    def get_threats(self) -> List[Threat]:
        self.signatures = self.parse_database(self.raw_data)
        self.threats = self.parse_threats(self.signatures)
        return self.threats

    @staticmethod
    def parse_database(data:bytes) -> List[BaseSig]:
        counter = defaultdict(int)
        ptr:int = 0
        entry:BaseSig
        database:List[BaseSig] = []

        # progress bar setup
        progress_bar = tqdm(
                total=len(data),
                unit='bytes',
                bar_format='{l_bar}{bar:20}{r_bar}',
                colour='green',
                desc="Parsing database",
                leave=False)

        # parse
        while ptr < len(data):
            sig_type = SIG_TYPES[data[ptr]]
            if sig_type == "SIGNATURE_TYPE_THREAT_BEGIN":
                entry = ThreatBeginSig(data,ptr)
            elif sig_type == "SIGNATURE_TYPE_THREAT_END":
                entry = ThreatEndSig(data,ptr)
            elif sig_type == "SIGNATURE_TYPE_PEHSTR":
                entry = HStrSig(data,ptr)
            elif sig_type in HSTR_EXT_SIGS:
                entry = HStrExtSig(data,ptr)
            else:
                entry = BaseSig(data,ptr)
            # check if the sig_data has valid data size
            if len(entry.sig_data) > 0:
                database.append(entry)
            counter[entry.sig_type] += 1
            ptr += entry.size + 4
            progress_bar.update(entry.size + 4)

        progress_bar.close()
        return database

    @staticmethod
    def parse_threats(database:List[BaseSig]) -> List[Threat]:
        # init threats
        threats = []
        # parse
        # progress bar setup
        progress_bar = tqdm(
                total=len(database),
                unit='threats',
                bar_format='{l_bar}{bar:20}{r_bar}',
                colour='green',
                desc="Parsing threats",
                leave=False)

        for sig in database:
            if sig.sig_type == "SIGNATURE_TYPE_THREAT_BEGIN":
                threat = Threat(sig)
                continue
            elif sig.sig_type == "SIGNATURE_TYPE_THREAT_END":
                threats.append(threat)
                threat = None
            if threat:
                threat.add_signature(sig)
            
            progress_bar.update(1)

        progress_bar.close()

        return threats

    @staticmethod
    def get_meta_info(path) -> Tuple[str,str]:
        try:
            with exiftool.ExifToolHelper() as ef:
                metadata = ef.get_metadata(path)[0]
                return metadata['EXE:OriginalFileName'],metadata['EXE:ProductVersion']
        except:
            # use pefile if exiftool is not present or available on the system 
            return parse_pe_meta_info(path)

    @staticmethod
    def extract_vdm_data(filepath) -> bytes:
        """
        Extracts VDM data from the given file.

        Args:
            filepath (str): The path to the file from which to extract VDM data.

        Returns:
            bytes: The decompressed VDM data.

        Raises:
            ValueError: If the file does not contain the expected resource signature or if the file format is invalid.
            FileNotFoundError: If the file does not exist.
            AssertionError: If the decompressed data size does not match the expected size.
        """
        if os.path.exists(filepath):
            data = open(filepath,"rb").read()
        else:
            raise FileNotFoundError(filepath)

        base = data.index(b"RMDX") # Look for resource signature
        if not base:
            raise ValueError(f"Invalid file format. {filepath}")
        
        offset, size = struct.unpack("II", data[base + 0x18: base + 0x20])
        decompressed_data = zlib.decompress(data[base + offset + 8:], -15)
        assert len(decompressed_data) == size # Ensure correctness
        return decompressed_data

    @staticmethod
    def apply_delta_patch(delta_blob:bytes, base_data:bytes)->bytes:
        results:List[bytes] = []

        ptr = 0
        unknown_dword1,unknown_dword2 = struct.unpack("II", delta_blob[ptr:ptr+8])
        ptr += 8

        while ptr < len(delta_blob):
            info = struct.unpack("H", delta_blob[ptr:ptr+2])[0]
            ptr += 2
            if info & 0x7fff == info:
                # append from delta 
                results.append(delta_blob[ptr:ptr+info])
                ptr += info
            else:
                # append from base
                base_offset = struct.unpack("I", delta_blob[ptr:ptr+4])[0]
                ptr += 4
                size = (info & 0x7fff) + 6
                results.append(base_data[base_offset:base_offset+size])
        return b"".join(results) # optimization / using "+=" for large bytes is too slow.  


from collections import defaultdict

THREAT_NAME_PREFIX = defaultdict(lambda:"")
THREAT_NAME_PREFIX[0x0185] = "Backdoor:BAT/"
THREAT_NAME_PREFIX[0x0186] = "Backdoor:FreeBSD/"
THREAT_NAME_PREFIX[0x0187] = "Backdoor:MSIL/"
THREAT_NAME_PREFIX[0x0189] = "TrojanDownloader:BAT/"
THREAT_NAME_PREFIX[0x018a] = "TrojanDownloader:FreeBSD/"
THREAT_NAME_PREFIX[0x018b] = "TrojanDownloader:MSIL/"
THREAT_NAME_PREFIX[0x018d] = "TrojanDropper:BAT/"
THREAT_NAME_PREFIX[0x018e] = "TrojanDropper:FreeBSD/"
THREAT_NAME_PREFIX[0x018f] = "TrojanDropper:MSIL/"
THREAT_NAME_PREFIX[0x0191] = "Spammer:BAT/"
THREAT_NAME_PREFIX[0x0192] = "Spammer:FreeBSD/"
THREAT_NAME_PREFIX[0x0193] = "Spammer:MSIL/"
THREAT_NAME_PREFIX[0x0195] = "DDoS:BAT/"
THREAT_NAME_PREFIX[0x0196] = "DDoS:FreeBSD/"
THREAT_NAME_PREFIX[0x0197] = "DDoS:MSIL/"
THREAT_NAME_PREFIX[0x0199] = "DoS:BAT/"
THREAT_NAME_PREFIX[0x019a] = "DoS:FreeBSD/"
THREAT_NAME_PREFIX[0x019b] = "DoS:MSIL/"
THREAT_NAME_PREFIX[0x019d] = "Joke:BAT/"
THREAT_NAME_PREFIX[0x019e] = "Joke:FreeBSD/"
THREAT_NAME_PREFIX[0x019f] = "Joke:MSIL/"
THREAT_NAME_PREFIX[0x01a1] = "PWS:BAT/"
THREAT_NAME_PREFIX[0x01a2] = "PWS:FreeBSD/"
THREAT_NAME_PREFIX[0x01a3] = "PWS:MSIL/"
THREAT_NAME_PREFIX[0x01a5] = "Worm:BAT/"
THREAT_NAME_PREFIX[0x01a6] = "Worm:FreeBSD/"
THREAT_NAME_PREFIX[0x01a7] = "Worm:MSIL/"
THREAT_NAME_PREFIX[0x01aa] = "Flooder:FreeBSD/"
THREAT_NAME_PREFIX[0x01ad] = "Trojan:BAT/"
THREAT_NAME_PREFIX[0x01ae] = "Trojan:FreeBSD/"
THREAT_NAME_PREFIX[0x01af] = "Trojan:MSIL/"
THREAT_NAME_PREFIX[0x01b1] = "Virus:BAT/"
THREAT_NAME_PREFIX[0x01b2] = "Virus:FreeBSD/"
THREAT_NAME_PREFIX[0x01b3] = "Virus:MSIL/"
THREAT_NAME_PREFIX[0x01b5] = "Constructor:BAT/"
THREAT_NAME_PREFIX[0x01b6] = "Constructor:FreeBSD/"
THREAT_NAME_PREFIX[0x01b7] = "Constructor:MSIL/"
THREAT_NAME_PREFIX[0x01ba] = "Nuker:FreeBSD/"
THREAT_NAME_PREFIX[0x01be] = "Spoofer:FreeBSD/"
THREAT_NAME_PREFIX[0x01c9] = "TrojanSpy:BAT/"
THREAT_NAME_PREFIX[0x01ca] = "TrojanSpy:FreeBSD/"
THREAT_NAME_PREFIX[0x01cb] = "TrojanSpy:MSIL/"
THREAT_NAME_PREFIX[0x01cd] = "VirTool:BAT/"
THREAT_NAME_PREFIX[0x01ce] = "VirTool:FreeBSD/"
THREAT_NAME_PREFIX[0x01cf] = "VirTool:MSIL/"
THREAT_NAME_PREFIX[0x01d1] = "Exploit:BAT/"
THREAT_NAME_PREFIX[0x01d2] = "Exploit:FreeBSD/"
THREAT_NAME_PREFIX[0x01d3] = "Exploit:MSIL/"
THREAT_NAME_PREFIX[0x01d5] = "TrojanClicker:BAT/"
THREAT_NAME_PREFIX[0x01d6] = "TrojanClicker:FreeBSD/"
THREAT_NAME_PREFIX[0x01d7] = "TrojanClicker:MSIL/"
THREAT_NAME_PREFIX[0x01d9] = "HackTool:BAT/"
THREAT_NAME_PREFIX[0x01da] = "HackTool:FreeBSD/"
THREAT_NAME_PREFIX[0x01db] = "HackTool:MSIL/"
THREAT_NAME_PREFIX[0x01dd] = "TrojanProxy:BAT/"
THREAT_NAME_PREFIX[0x01de] = "TrojanProxy:FreeBSD/"
THREAT_NAME_PREFIX[0x01df] = "TrojanProxy:MSIL/"
THREAT_NAME_PREFIX[0x01ea] = "TrojanNotifier:FreeBSD/"
THREAT_NAME_PREFIX[0x03a5] = "Worm:BAT/"
THREAT_NAME_PREFIX[0x03ad] = "Trojan:BAT/"
THREAT_NAME_PREFIX[0x03b1] = "Virus:BAT/"
THREAT_NAME_PREFIX[0x09a5] = "Worm:BAT/"
THREAT_NAME_PREFIX[0x09a7] = "Worm:MSIL/"
THREAT_NAME_PREFIX[0x09b1] = "Virus:BAT/"
THREAT_NAME_PREFIX[0x1187] = "Backdoor:MSIL/"
THREAT_NAME_PREFIX[0x1189] = "TrojanDownloader:BAT/"
THREAT_NAME_PREFIX[0x118f] = "TrojanDropper:MSIL/"
THREAT_NAME_PREFIX[0x11ad] = "Trojan:BAT/"
THREAT_NAME_PREFIX[0x11b1] = "Virus:BAT/"
THREAT_NAME_PREFIX[0x11cf] = "VirTool:MSIL/"
THREAT_NAME_PREFIX[0x2184] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x2185] = "Backdoor:VBS/"
THREAT_NAME_PREFIX[0x2186] = "Backdoor:OS2/"
THREAT_NAME_PREFIX[0x2187] = "Backdoor:INF/"
THREAT_NAME_PREFIX[0x2188] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x2189] = "TrojanDownloader:VBS/"
THREAT_NAME_PREFIX[0x218c] = "TrojanDropper:Win32/"
THREAT_NAME_PREFIX[0x218d] = "TrojanDropper:VBS/"
THREAT_NAME_PREFIX[0x2190] = "Spammer:Win32/"
THREAT_NAME_PREFIX[0x2191] = "Spammer:VBS/"
THREAT_NAME_PREFIX[0x2194] = "DDoS:Win32/"
THREAT_NAME_PREFIX[0x2195] = "DDoS:VBS/"
THREAT_NAME_PREFIX[0x2198] = "DoS:Win32/"
THREAT_NAME_PREFIX[0x219c] = "Joke:Win32/"
THREAT_NAME_PREFIX[0x219d] = "Joke:VBS/"
THREAT_NAME_PREFIX[0x219f] = "Joke:INF/"
THREAT_NAME_PREFIX[0x21a0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x21a1] = "PWS:VBS/"
THREAT_NAME_PREFIX[0x21a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x21a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x21a7] = "Worm:INF/"
THREAT_NAME_PREFIX[0x21a8] = "Flooder:Win32/"
THREAT_NAME_PREFIX[0x21ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x21ad] = "Trojan:VBS/"
THREAT_NAME_PREFIX[0x21ae] = "Trojan:OS2/"
THREAT_NAME_PREFIX[0x21af] = "Trojan:INF/"
THREAT_NAME_PREFIX[0x21b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x21b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x21b2] = "Virus:OS2/"
THREAT_NAME_PREFIX[0x21b3] = "Virus:INF/"
THREAT_NAME_PREFIX[0x21b4] = "Constructor:Win32/"
THREAT_NAME_PREFIX[0x21b5] = "Constructor:VBS/"
THREAT_NAME_PREFIX[0x21b8] = "Nuker:Win32/"
THREAT_NAME_PREFIX[0x21bc] = "Spoofer:Win32/"
THREAT_NAME_PREFIX[0x21c0] = "Tool:Win32/"
THREAT_NAME_PREFIX[0x21c8] = "TrojanSpy:Win32/"
THREAT_NAME_PREFIX[0x21c9] = "TrojanSpy:VBS/"
THREAT_NAME_PREFIX[0x21cc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x21cd] = "VirTool:VBS/"
THREAT_NAME_PREFIX[0x21cf] = "VirTool:INF/"
THREAT_NAME_PREFIX[0x21d0] = "Exploit:Win32/"
THREAT_NAME_PREFIX[0x21d1] = "Exploit:VBS/"
THREAT_NAME_PREFIX[0x21d3] = "Exploit:INF/"
THREAT_NAME_PREFIX[0x21d4] = "TrojanClicker:Win32/"
THREAT_NAME_PREFIX[0x21d5] = "TrojanClicker:VBS/"
THREAT_NAME_PREFIX[0x21d8] = "HackTool:Win32/"
THREAT_NAME_PREFIX[0x21d9] = "HackTool:VBS/"
THREAT_NAME_PREFIX[0x21dc] = "TrojanProxy:Win32/"
THREAT_NAME_PREFIX[0x21dd] = "TrojanProxy:VBS/"
THREAT_NAME_PREFIX[0x21e8] = "TrojanNotifier:Win32/"
THREAT_NAME_PREFIX[0x2384] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x2388] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x238c] = "TrojanDropper:Win32/"
THREAT_NAME_PREFIX[0x238d] = "TrojanDropper:VBS/"
THREAT_NAME_PREFIX[0x2390] = "Spammer:Win32/"
THREAT_NAME_PREFIX[0x2394] = "DDoS:Win32/"
THREAT_NAME_PREFIX[0x239c] = "Joke:Win32/"
THREAT_NAME_PREFIX[0x23a0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x23a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x23a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x23ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x23b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x23b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x23c8] = "TrojanSpy:Win32/"
THREAT_NAME_PREFIX[0x23cc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x23d4] = "TrojanClicker:Win32/"
THREAT_NAME_PREFIX[0x23d8] = "HackTool:Win32/"
THREAT_NAME_PREFIX[0x23dc] = "TrojanProxy:Win32/"
THREAT_NAME_PREFIX[0x2584] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x25a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x25a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x25ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x25b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x25b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x2788] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x27a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x27ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x27b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x29a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x29a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x29b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x29b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x2b84] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x2b88] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x2b8c] = "TrojanDropper:Win32/"
THREAT_NAME_PREFIX[0x2b90] = "Spammer:Win32/"
THREAT_NAME_PREFIX[0x2b98] = "DoS:Win32/"
THREAT_NAME_PREFIX[0x2ba0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x2ba4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x2ba5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x2bac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x2bb0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x2bb1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x2bc8] = "TrojanSpy:Win32/"
THREAT_NAME_PREFIX[0x2bcc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x2bd0] = "Exploit:Win32/"
THREAT_NAME_PREFIX[0x2bdc] = "TrojanProxy:Win32/"
THREAT_NAME_PREFIX[0x2d84] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x2da4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x2dac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x2dcc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x2fac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x2fb0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x3184] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x3185] = "Backdoor:VBS/"
THREAT_NAME_PREFIX[0x3188] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x3189] = "TrojanDownloader:VBS/"
THREAT_NAME_PREFIX[0x318c] = "TrojanDropper:Win32/"
THREAT_NAME_PREFIX[0x318d] = "TrojanDropper:VBS/"
THREAT_NAME_PREFIX[0x3190] = "Spammer:Win32/"
THREAT_NAME_PREFIX[0x31a0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x31a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x31a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x31a7] = "Worm:INF/"
THREAT_NAME_PREFIX[0x31ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x31ad] = "Trojan:VBS/"
THREAT_NAME_PREFIX[0x31b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x31b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x31b5] = "Constructor:VBS/"
THREAT_NAME_PREFIX[0x31c8] = "TrojanSpy:Win32/"
THREAT_NAME_PREFIX[0x31cc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x31cf] = "VirTool:INF/"
THREAT_NAME_PREFIX[0x31d0] = "Exploit:Win32/"
THREAT_NAME_PREFIX[0x31d1] = "Exploit:VBS/"
THREAT_NAME_PREFIX[0x31d4] = "TrojanClicker:Win32/"
THREAT_NAME_PREFIX[0x31d8] = "HackTool:Win32/"
THREAT_NAME_PREFIX[0x31dc] = "TrojanProxy:Win32/"
THREAT_NAME_PREFIX[0x33a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x3584] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x3588] = "TrojanDownloader:Win32/"
THREAT_NAME_PREFIX[0x358c] = "TrojanDropper:Win32/"
THREAT_NAME_PREFIX[0x3590] = "Spammer:Win32/"
THREAT_NAME_PREFIX[0x35a0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x35a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x35ac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x35b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x35b4] = "Constructor:Win32/"
THREAT_NAME_PREFIX[0x35c8] = "TrojanSpy:Win32/"
THREAT_NAME_PREFIX[0x35cc] = "VirTool:Win32/"
THREAT_NAME_PREFIX[0x35d4] = "TrojanClicker:Win32/"
THREAT_NAME_PREFIX[0x35d8] = "HackTool:Win32/"
THREAT_NAME_PREFIX[0x35dc] = "TrojanProxy:Win32/"
THREAT_NAME_PREFIX[0x37a4] = "Worm:Win32/"
THREAT_NAME_PREFIX[0x37a5] = "Worm:VBS/"
THREAT_NAME_PREFIX[0x37b0] = "Virus:Win32/"
THREAT_NAME_PREFIX[0x37b1] = "Virus:VBS/"
THREAT_NAME_PREFIX[0x39a0] = "PWS:Win32/"
THREAT_NAME_PREFIX[0x3b84] = "Backdoor:Win32/"
THREAT_NAME_PREFIX[0x3bac] = "Trojan:Win32/"
THREAT_NAME_PREFIX[0x4184] = "Backdoor:Win95/"
THREAT_NAME_PREFIX[0x4185] = "Backdoor:JS/"
THREAT_NAME_PREFIX[0x4189] = "TrojanDownloader:JS/"
THREAT_NAME_PREFIX[0x418d] = "TrojanDropper:JS/"
THREAT_NAME_PREFIX[0x4191] = "Spammer:JS/"
THREAT_NAME_PREFIX[0x4199] = "DoS:JS/"
THREAT_NAME_PREFIX[0x419d] = "Joke:JS/"
THREAT_NAME_PREFIX[0x41a0] = "PWS:Win95/"
THREAT_NAME_PREFIX[0x41a1] = "PWS:JS/"
THREAT_NAME_PREFIX[0x41a4] = "Worm:Win95/"
THREAT_NAME_PREFIX[0x41a5] = "Worm:JS/"
THREAT_NAME_PREFIX[0x41ac] = "Trojan:Win95/"
THREAT_NAME_PREFIX[0x41ad] = "Trojan:JS/"
THREAT_NAME_PREFIX[0x41ae] = "Trojan:Win98/"
THREAT_NAME_PREFIX[0x41af] = "Trojan:SunOS/"
THREAT_NAME_PREFIX[0x41b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x41b1] = "Virus:JS/"
THREAT_NAME_PREFIX[0x41b5] = "Constructor:JS/"
THREAT_NAME_PREFIX[0x41c9] = "TrojanSpy:JS/"
THREAT_NAME_PREFIX[0x41cc] = "VirTool:Win95/"
THREAT_NAME_PREFIX[0x41cd] = "VirTool:JS/"
THREAT_NAME_PREFIX[0x41d1] = "Exploit:JS/"
THREAT_NAME_PREFIX[0x41d3] = "Exploit:SunOS/"
THREAT_NAME_PREFIX[0x41d5] = "TrojanClicker:JS/"
THREAT_NAME_PREFIX[0x41d9] = "HackTool:JS/"
THREAT_NAME_PREFIX[0x41dd] = "TrojanProxy:JS/"
THREAT_NAME_PREFIX[0x41df] = "TrojanProxy:SunOS/"
THREAT_NAME_PREFIX[0x43a5] = "Worm:JS/"
THREAT_NAME_PREFIX[0x43ad] = "Trojan:JS/"
THREAT_NAME_PREFIX[0x43b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x43b1] = "Virus:JS/"
THREAT_NAME_PREFIX[0x45a5] = "Worm:JS/"
THREAT_NAME_PREFIX[0x45b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x47b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x49a5] = "Worm:JS/"
THREAT_NAME_PREFIX[0x49ad] = "Trojan:JS/"
THREAT_NAME_PREFIX[0x49b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x4b8d] = "TrojanDropper:JS/"
THREAT_NAME_PREFIX[0x4bb0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x4bb1] = "Virus:JS/"
THREAT_NAME_PREFIX[0x4db0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x5189] = "TrojanDownloader:JS/"
THREAT_NAME_PREFIX[0x5199] = "DoS:JS/"
THREAT_NAME_PREFIX[0x51ad] = "Trojan:JS/"
THREAT_NAME_PREFIX[0x51b0] = "Virus:Win95/"
THREAT_NAME_PREFIX[0x51d1] = "Exploit:JS/"
THREAT_NAME_PREFIX[0x55a0] = "PWS:Win95/"
THREAT_NAME_PREFIX[0x57ad] = "Trojan:JS/"
THREAT_NAME_PREFIX[0x6184] = "Backdoor:WinNT/"
THREAT_NAME_PREFIX[0x6185] = "Backdoor:Java/"
THREAT_NAME_PREFIX[0x6186] = "Backdoor:Win2K/"
THREAT_NAME_PREFIX[0x6188] = "TrojanDownloader:WinNT/"
THREAT_NAME_PREFIX[0x6189] = "TrojanDownloader:Java/"
THREAT_NAME_PREFIX[0x618c] = "TrojanDropper:WinNT/"
THREAT_NAME_PREFIX[0x618d] = "TrojanDropper:Java/"
THREAT_NAME_PREFIX[0x6190] = "Spammer:WinNT/"
THREAT_NAME_PREFIX[0x6191] = "Spammer:Java/"
THREAT_NAME_PREFIX[0x6199] = "DoS:Java/"
THREAT_NAME_PREFIX[0x619d] = "Joke:Java/"
THREAT_NAME_PREFIX[0x61a0] = "PWS:WinNT/"
THREAT_NAME_PREFIX[0x61a1] = "PWS:Java/"
THREAT_NAME_PREFIX[0x61a4] = "Worm:WinNT/"
THREAT_NAME_PREFIX[0x61a5] = "Worm:Java/"
THREAT_NAME_PREFIX[0x61a6] = "Worm:Win2K/"
THREAT_NAME_PREFIX[0x61a9] = "Flooder:Java/"
THREAT_NAME_PREFIX[0x61ac] = "Trojan:WinNT/"
THREAT_NAME_PREFIX[0x61ad] = "Trojan:Java/"
THREAT_NAME_PREFIX[0x61ae] = "Trojan:Win2K/"
THREAT_NAME_PREFIX[0x61af] = "Trojan:Netware/"
THREAT_NAME_PREFIX[0x61b0] = "Virus:WinNT/"
THREAT_NAME_PREFIX[0x61b1] = "Virus:Java/"
THREAT_NAME_PREFIX[0x61b2] = "Virus:Win2K/"
THREAT_NAME_PREFIX[0x61b5] = "Constructor:Java/"
THREAT_NAME_PREFIX[0x61c8] = "TrojanSpy:WinNT/"
THREAT_NAME_PREFIX[0x61c9] = "TrojanSpy:Java/"
THREAT_NAME_PREFIX[0x61cc] = "VirTool:WinNT/"
THREAT_NAME_PREFIX[0x61cd] = "VirTool:Java/"
THREAT_NAME_PREFIX[0x61d0] = "Exploit:WinNT/"
THREAT_NAME_PREFIX[0x61d1] = "Exploit:Java/"
THREAT_NAME_PREFIX[0x61d8] = "HackTool:WinNT/"
THREAT_NAME_PREFIX[0x61d9] = "HackTool:Java/"
THREAT_NAME_PREFIX[0x61dd] = "TrojanProxy:Java/"
THREAT_NAME_PREFIX[0x6384] = "Backdoor:WinNT/"
THREAT_NAME_PREFIX[0x63ac] = "Trojan:WinNT/"
THREAT_NAME_PREFIX[0x63b2] = "Virus:Win2K/"
THREAT_NAME_PREFIX[0x63cc] = "VirTool:WinNT/"
THREAT_NAME_PREFIX[0x69a5] = "Worm:Java/"
THREAT_NAME_PREFIX[0x69ad] = "Trojan:Java/"
THREAT_NAME_PREFIX[0x7184] = "Backdoor:WinNT/"
THREAT_NAME_PREFIX[0x7190] = "Spammer:WinNT/"
THREAT_NAME_PREFIX[0x71ac] = "Trojan:WinNT/"
THREAT_NAME_PREFIX[0x71ad] = "Trojan:Java/"
THREAT_NAME_PREFIX[0x71cc] = "VirTool:WinNT/"
THREAT_NAME_PREFIX[0x8184] = "Backdoor:DOS/"
THREAT_NAME_PREFIX[0x8185] = "Backdoor:IRC/"
THREAT_NAME_PREFIX[0x8186] = "Backdoor:AutoIt/"
THREAT_NAME_PREFIX[0x8188] = "TrojanDownloader:DOS/"
THREAT_NAME_PREFIX[0x818a] = "TrojanDownloader:AutoIt/"
THREAT_NAME_PREFIX[0x818c] = "TrojanDropper:DOS/"
THREAT_NAME_PREFIX[0x818e] = "TrojanDropper:AutoIt/"
THREAT_NAME_PREFIX[0x8190] = "Spammer:DOS/"
THREAT_NAME_PREFIX[0x8191] = "Spammer:IRC/"
THREAT_NAME_PREFIX[0x8194] = "DDoS:DOS/"
THREAT_NAME_PREFIX[0x8195] = "DDoS:IRC/"
THREAT_NAME_PREFIX[0x8198] = "DoS:DOS/"
THREAT_NAME_PREFIX[0x8199] = "DoS:IRC/"
THREAT_NAME_PREFIX[0x819c] = "Joke:DOS/"
THREAT_NAME_PREFIX[0x81a0] = "PWS:DOS/"
THREAT_NAME_PREFIX[0x81a2] = "PWS:AutoIt/"
THREAT_NAME_PREFIX[0x81a4] = "Worm:DOS/"
THREAT_NAME_PREFIX[0x81a5] = "Worm:IRC/"
THREAT_NAME_PREFIX[0x81a6] = "Worm:AutoIt/"
THREAT_NAME_PREFIX[0x81ac] = "Trojan:DOS/"
THREAT_NAME_PREFIX[0x81ad] = "Trojan:IRC/"
THREAT_NAME_PREFIX[0x81ae] = "Trojan:AutoIt/"
THREAT_NAME_PREFIX[0x81af] = "Trojan:DOS32/"
THREAT_NAME_PREFIX[0x81b0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x81b2] = "Virus:AutoIt/"
THREAT_NAME_PREFIX[0x81b4] = "Constructor:DOS/"
THREAT_NAME_PREFIX[0x81b5] = "Constructor:IRC/"
THREAT_NAME_PREFIX[0x81c8] = "TrojanSpy:DOS/"
THREAT_NAME_PREFIX[0x81ca] = "TrojanSpy:AutoIt/"
THREAT_NAME_PREFIX[0x81cc] = "VirTool:DOS/"
THREAT_NAME_PREFIX[0x81ce] = "VirTool:AutoIt/"
THREAT_NAME_PREFIX[0x81d0] = "Exploit:DOS/"
THREAT_NAME_PREFIX[0x81d1] = "Exploit:IRC/"
THREAT_NAME_PREFIX[0x81d2] = "Exploit:AutoIt/"
THREAT_NAME_PREFIX[0x81d6] = "TrojanClicker:AutoIt/"
THREAT_NAME_PREFIX[0x81d8] = "HackTool:DOS/"
THREAT_NAME_PREFIX[0x81d9] = "HackTool:IRC/"
THREAT_NAME_PREFIX[0x81da] = "HackTool:AutoIt/"
THREAT_NAME_PREFIX[0x81de] = "TrojanProxy:AutoIt/"
THREAT_NAME_PREFIX[0x8385] = "Backdoor:IRC/"
THREAT_NAME_PREFIX[0x83a4] = "Worm:DOS/"
THREAT_NAME_PREFIX[0x83ac] = "Trojan:DOS/"
THREAT_NAME_PREFIX[0x83ad] = "Trojan:IRC/"
THREAT_NAME_PREFIX[0x83b0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x85a4] = "Worm:DOS/"
THREAT_NAME_PREFIX[0x85b0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x87ac] = "Trojan:DOS/"
THREAT_NAME_PREFIX[0x87b0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x89a4] = "Worm:DOS/"
THREAT_NAME_PREFIX[0x8ba5] = "Worm:IRC/"
THREAT_NAME_PREFIX[0x8bb0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x9185] = "Backdoor:IRC/"
THREAT_NAME_PREFIX[0x91a4] = "Worm:DOS/"
THREAT_NAME_PREFIX[0x91ad] = "Trojan:IRC/"
THREAT_NAME_PREFIX[0x91b0] = "Virus:DOS/"
THREAT_NAME_PREFIX[0x9b85] = "Backdoor:IRC/"
THREAT_NAME_PREFIX[0xa184] = "Backdoor:Linux/"
THREAT_NAME_PREFIX[0xa186] = "Backdoor:WinCE/"
THREAT_NAME_PREFIX[0xa187] = "Backdoor:MacOS_X/"
THREAT_NAME_PREFIX[0xa188] = "TrojanDownloader:Linux/"
THREAT_NAME_PREFIX[0xa189] = "TrojanDownloader:W97M/"
THREAT_NAME_PREFIX[0xa18a] = "TrojanDownloader:WinCE/"
THREAT_NAME_PREFIX[0xa18b] = "TrojanDownloader:MacOS_X/"
THREAT_NAME_PREFIX[0xa18c] = "TrojanDropper:Linux/"
THREAT_NAME_PREFIX[0xa18d] = "TrojanDropper:W97M/"
THREAT_NAME_PREFIX[0xa18f] = "TrojanDropper:MacOS_X/"
THREAT_NAME_PREFIX[0xa190] = "Spammer:Linux/"
THREAT_NAME_PREFIX[0xa193] = "Spammer:MacOS_X/"
THREAT_NAME_PREFIX[0xa194] = "DDoS:Linux/"
THREAT_NAME_PREFIX[0xa197] = "DDoS:MacOS_X/"
THREAT_NAME_PREFIX[0xa198] = "DoS:Linux/"
THREAT_NAME_PREFIX[0xa19b] = "DoS:MacOS_X/"
THREAT_NAME_PREFIX[0xa19c] = "Joke:Linux/"
THREAT_NAME_PREFIX[0xa1a0] = "PWS:Linux/"
THREAT_NAME_PREFIX[0xa1a1] = "PWS:W97M/"
THREAT_NAME_PREFIX[0xa1a3] = "PWS:MacOS_X/"
THREAT_NAME_PREFIX[0xa1a4] = "Worm:Linux/"
THREAT_NAME_PREFIX[0xa1a6] = "Worm:WinCE/"
THREAT_NAME_PREFIX[0xa1a7] = "Worm:MacOS_X/"
THREAT_NAME_PREFIX[0xa1a8] = "Flooder:Linux/"
THREAT_NAME_PREFIX[0xa1ac] = "Trojan:Linux/"
THREAT_NAME_PREFIX[0xa1ad] = "Trojan:W97M/"
THREAT_NAME_PREFIX[0xa1ae] = "Trojan:WinCE/"
THREAT_NAME_PREFIX[0xa1af] = "Trojan:MacOS_X/"
THREAT_NAME_PREFIX[0xa1b0] = "Virus:Linux/"
THREAT_NAME_PREFIX[0xa1b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xa1b2] = "Virus:WinCE/"
THREAT_NAME_PREFIX[0xa1b3] = "Virus:MacOS_X/"
THREAT_NAME_PREFIX[0xa1b4] = "Constructor:Linux/"
THREAT_NAME_PREFIX[0xa1b5] = "Constructor:W97M/"
THREAT_NAME_PREFIX[0xa1b8] = "Nuker:Linux/"
THREAT_NAME_PREFIX[0xa1bc] = "Spoofer:Linux/"
THREAT_NAME_PREFIX[0xa1c8] = "TrojanSpy:Linux/"
THREAT_NAME_PREFIX[0xa1cb] = "TrojanSpy:MacOS_X/"
THREAT_NAME_PREFIX[0xa1cc] = "VirTool:Linux/"
THREAT_NAME_PREFIX[0xa1cf] = "VirTool:MacOS_X/"
THREAT_NAME_PREFIX[0xa1d0] = "Exploit:Linux/"
THREAT_NAME_PREFIX[0xa1d1] = "Exploit:W97M/"
THREAT_NAME_PREFIX[0xa1d3] = "Exploit:MacOS_X/"
THREAT_NAME_PREFIX[0xa1d4] = "TrojanClicker:Linux/"
THREAT_NAME_PREFIX[0xa1d8] = "HackTool:Linux/"
THREAT_NAME_PREFIX[0xa1d9] = "HackTool:W97M/"
THREAT_NAME_PREFIX[0xa1db] = "HackTool:MacOS_X/"
THREAT_NAME_PREFIX[0xa1dc] = "TrojanProxy:Linux/"
THREAT_NAME_PREFIX[0xa1e8] = "TrojanNotifier:Linux/"
THREAT_NAME_PREFIX[0xa3ac] = "Trojan:Linux/"
THREAT_NAME_PREFIX[0xa3b0] = "Virus:Linux/"
THREAT_NAME_PREFIX[0xa3b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xa3b2] = "Virus:WinCE/"
THREAT_NAME_PREFIX[0xa5b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xa7b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xa9b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xabb1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xad84] = "Backdoor:Linux/"
THREAT_NAME_PREFIX[0xb1b0] = "Virus:Linux/"
THREAT_NAME_PREFIX[0xb1b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xb1d0] = "Exploit:Linux/"
THREAT_NAME_PREFIX[0xb1d8] = "HackTool:Linux/"
THREAT_NAME_PREFIX[0xb9b1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xbbb1] = "Virus:W97M/"
THREAT_NAME_PREFIX[0xc184] = "Backdoor:Win16/"
THREAT_NAME_PREFIX[0xc185] = "Backdoor:X97M/"
THREAT_NAME_PREFIX[0xc189] = "TrojanDownloader:X97M/"
THREAT_NAME_PREFIX[0xc18a] = "TrojanDownloader:SymbOS/"
THREAT_NAME_PREFIX[0xc18c] = "TrojanDropper:Win16/"
THREAT_NAME_PREFIX[0xc18d] = "TrojanDropper:X97M/"
THREAT_NAME_PREFIX[0xc18e] = "TrojanDropper:SymbOS/"
THREAT_NAME_PREFIX[0xc190] = "Spammer:Win16/"
THREAT_NAME_PREFIX[0xc194] = "DDoS:Win16/"
THREAT_NAME_PREFIX[0xc198] = "DoS:Win16/"
THREAT_NAME_PREFIX[0xc19c] = "Joke:Win16/"
THREAT_NAME_PREFIX[0xc1a0] = "PWS:Win16/"
THREAT_NAME_PREFIX[0xc1a1] = "PWS:X97M/"
THREAT_NAME_PREFIX[0xc1a4] = "Worm:Win16/"
THREAT_NAME_PREFIX[0xc1a5] = "Worm:X97M/"
THREAT_NAME_PREFIX[0xc1a6] = "Worm:SymbOS/"
THREAT_NAME_PREFIX[0xc1ac] = "Trojan:Win16/"
THREAT_NAME_PREFIX[0xc1ad] = "Trojan:X97M/"
THREAT_NAME_PREFIX[0xc1ae] = "Trojan:SymbOS/"
THREAT_NAME_PREFIX[0xc1af] = "Trojan:AppleScript/"
THREAT_NAME_PREFIX[0xc1b0] = "Virus:Win16/"
THREAT_NAME_PREFIX[0xc1b1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xc1b4] = "Constructor:Win16/"
THREAT_NAME_PREFIX[0xc1b8] = "Nuker:Win16/"
THREAT_NAME_PREFIX[0xc1c8] = "TrojanSpy:Win16/"
THREAT_NAME_PREFIX[0xc1c9] = "TrojanSpy:X97M/"
THREAT_NAME_PREFIX[0xc1ca] = "TrojanSpy:SymbOS/"
THREAT_NAME_PREFIX[0xc1cd] = "VirTool:X97M/"
THREAT_NAME_PREFIX[0xc1d1] = "Exploit:X97M/"
THREAT_NAME_PREFIX[0xc1d8] = "HackTool:Win16/"
THREAT_NAME_PREFIX[0xc3ac] = "Trojan:Win16/"
THREAT_NAME_PREFIX[0xc3b1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xc5b1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xc7b1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xc9a4] = "Worm:Win16/"
THREAT_NAME_PREFIX[0xc9af] = "Trojan:AppleScript/"
THREAT_NAME_PREFIX[0xc9b1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xcbb0] = "Virus:Win16/"
THREAT_NAME_PREFIX[0xcbb1] = "Virus:X97M/"
THREAT_NAME_PREFIX[0xd1b0] = "Virus:Win16/"
THREAT_NAME_PREFIX[0xe184] = "Backdoor:MacOS/"
THREAT_NAME_PREFIX[0xe188] = "TrojanDownloader:MacOS/"
THREAT_NAME_PREFIX[0xe18c] = "TrojanDropper:MacOS/"
THREAT_NAME_PREFIX[0xe18e] = "TrojanDropper:WinHLP/"
THREAT_NAME_PREFIX[0xe190] = "Spammer:MacOS/"
THREAT_NAME_PREFIX[0xe194] = "DDoS:MacOS/"
THREAT_NAME_PREFIX[0xe198] = "DoS:MacOS/"
THREAT_NAME_PREFIX[0xe19c] = "Joke:MacOS/"
THREAT_NAME_PREFIX[0xe1a0] = "PWS:MacOS/"
THREAT_NAME_PREFIX[0xe1a4] = "Worm:MacOS/"
THREAT_NAME_PREFIX[0xe1a8] = "Flooder:MacOS/"
THREAT_NAME_PREFIX[0xe1ac] = "Trojan:MacOS/"
THREAT_NAME_PREFIX[0xe1ad] = "Trojan:PP97M/"
THREAT_NAME_PREFIX[0xe1ae] = "Trojan:WinHLP/"
THREAT_NAME_PREFIX[0xe1b0] = "Virus:MacOS/"
THREAT_NAME_PREFIX[0xe1b1] = "Virus:PP97M/"
THREAT_NAME_PREFIX[0xe1b2] = "Virus:WinHLP/"
THREAT_NAME_PREFIX[0xe1b4] = "Constructor:MacOS/"
THREAT_NAME_PREFIX[0xe1b5] = "Constructor:PP97M/"
THREAT_NAME_PREFIX[0xe1b8] = "Nuker:MacOS/"
THREAT_NAME_PREFIX[0xe1bc] = "Spoofer:MacOS/"
THREAT_NAME_PREFIX[0xe1c8] = "TrojanSpy:MacOS/"
THREAT_NAME_PREFIX[0xe1cc] = "VirTool:MacOS/"
THREAT_NAME_PREFIX[0xe1d0] = "Exploit:MacOS/"
THREAT_NAME_PREFIX[0xe1d4] = "TrojanClicker:MacOS/"
THREAT_NAME_PREFIX[0xe1d8] = "HackTool:MacOS/"
THREAT_NAME_PREFIX[0xe1dc] = "TrojanProxy:MacOS/"
THREAT_NAME_PREFIX[0xe1e8] = "TrojanNotifier:MacOS/"
THREAT_NAME_PREFIX[0xe3b2] = "Virus:WinHLP/"
THREAT_NAME_PREFIX[0xe5b1] = "Virus:PP97M/"
THREAT_NAME_PREFIX[0xe9b0] = "Virus:MacOS/"
THREAT_NAME_PREFIX[0xebb1] = "Virus:PP97M/"


from typing import Union
import re

def hexdump(data, length=16):
    """
    Display a hexadecimal dump of byte data.

    Args:
        data (bytes): The byte data to be displayed.
        length (int): The number of bytes per line (default is 16).
    """
    def format_line(addr, line):
        hex_part = ' '.join(f'{byte:02x}' for byte in line)
        ascii_part = ''.join(chr(byte) if 32 <= byte < 127 else '.' for byte in line)
        return f'{addr:08x}  {hex_part:<{length * 3}}  {ascii_part}'

    for i in range(0, len(data), length):
        line = data[i:i + length]
        print(format_line(i, line))


def is_printable_ascii(data: Union[bytes,str]) -> bool:
    """
    Check if all characters in the byte data are printable ASCII characters.

    Args:
        data (bytes or str): The byte data to check.

    Returns:
        bool: True if all characters are printable ASCII, False otherwise.
    """
    if isinstance(data,bytes):
        return all(32 <= c <= 126 for c in data)
    elif isinstance(data,str):
        return all(32 <= ord(c) <= 126 for c in data)
    else:
        raise ValueError("Not supported type.")


def is_printable_utf16_le_ascii(byte_data:bytes) -> bool:
    """
    Check if the given byte data is encoded as UTF-16-LE and contains only ASCII characters.

    Args:
        byte_data (bytes): The byte data to check.

    Returns:
        bool: True if the byte data is UTF-16-LE encoded and contains only ASCII characters, False otherwise.
    """
    if len(byte_data) % 2 != 0:
        # Length of byte_data must be even for valid UTF-16 encoding
        return False
    try:
        # Decode the byte data using UTF-16-LE
        decoded_str = byte_data.decode('utf-16-le')
        
        # Check if all characters are ASCII
        return all(32 <= ord(c) <= 126 for c in decoded_str)
    except UnicodeDecodeError:
        # If decoding fails, it's not a valid UTF-16-LE encoding
        return False


def is_ascii(c:int) -> bool:
    """
    Check if the given integer value corresponds to an ASCII character.

    Args:
        c (int): The integer value to check.

    Returns:
        bool: True if the integer value corresponds to an ASCII character, False otherwise.
    """
    if c < 128:
        return True
    else:
        return False


def all_elements_equal(lst):
    """Check if all elements in the list are equal.

    Args:
        lst (list): A list of elements to be checked.

    Returns:
        bool: True if all elements are equal or the list is empty, False otherwise.
        
    Examples:
        >>> all_elements_equal([1, 1, 1, 1])
        True
        >>> all_elements_equal([1, 2, 1, 1])
        False
        >>> all_elements_equal([])
        True
    """
    if not lst: 
        return True
    first_element = lst[0]
    return all(element == first_element for element in lst)


def is_validate_filesize(input_string):
    pattern = r'^(\d+)\s*(KB|MB|)$'
    match = re.match(pattern, input_string.strip(), re.IGNORECASE)
    if match:
        value, unit = match.groups()
        try:
            value = int(value)
            if unit.upper() == 'KB':
                value *= 1024
            elif unit.upper() == 'MB':
                value *= 1024 * 1024
            if value > 0:
                return True
            else:
                return False
        except ValueError:
            return False
    else:
        return False



from typing import Tuple,List
import struct
from .base import BaseSig
from defender2yara.defender.subrule.hstr import HStrSubRule, HStrExtSubRule
from defender2yara.defender.constant import SIG_TYPES


HSTR_EXT_SIGS = [
    "SIGNATURE_TYPE_PEHSTR_EXT",
    "SIGNATURE_TYPE_ELFHSTR_EXT",
    "SIGNATURE_TYPE_MACHOHSTR_EXT",
    "SIGNATURE_TYPE_DOSHSTR_EXT",
    "SIGNATURE_TYPE_MACROHSTR_EXT",
    "SIGNATURE_TYPE_DEXHSTR_EXT",
    "SIGNATURE_TYPE_JAVAHSTR_EXT",
    "SIGNATURE_TYPE_ARHSTR_EXT",
    "SIGNATURE_TYPE_SWFHSTR_EXT",
    "SIGNATURE_TYPE_AUTOITHSTR_EXT",
    "SIGNATURE_TYPE_INNOHSTR_EXT",
    "SIGNATURE_TYPE_CMDHSTR_EXT",
    "SIGNATURE_TYPE_MDBHSTR_EXT",
    "SIGNATURE_TYPE_DMGHSTR_EXT"
]


class HStrSig(BaseSig):
    def __init__(self,data:bytes,ptr:int=0) -> None:
        # parse common header
        self.parse_common_header(data,ptr)
        assert self.sig_type == SIG_TYPES[0x61]

        # parse signature data header
        self.unknown_u16_1 = struct.unpack('<H',self.sig_data[0:2])[0]
        self.threshold = struct.unpack('<H',self.sig_data[2:4])[0]
        self.rule_numbers = struct.unpack('<H',self.sig_data[4:6])[0]
        self.unknown_u8_1 = self.sig_data[6]

        # parse subrules
        self.subrules = []
        pos = 7
        while pos < len(self.sig_data):
            subrule = HStrSubRule(self.sig_data,pos)
            # no more subrule
            if subrule.rule_weight == 0:
                break
            self.subrules.append(subrule)
            pos += 3 + subrule.rule_size


class HStrExtSig(BaseSig):
    def __init__(self,data:bytes,ptr:int=0) -> None:
        # parse common header
        self.parse_common_header(data,ptr)
        assert self.sig_type in HSTR_EXT_SIGS

        # parse signature data header
        self.unknown_u16_1 = struct.unpack('<H',self.sig_data[0:2])[0]
        self.threshold = struct.unpack('<H',self.sig_data[2:4])[0]
        self.rule_numbers = struct.unpack('<H',self.sig_data[4:6])[0]
        self.unknown_u8_1 = self.sig_data[6]

        # parse subrules 
        self.subrules = []
        ptr = 7
        while ptr < len(self.sig_data):
            subrule = HStrExtSubRule(self.sig_data,ptr)
            # no more subrule
            if subrule.rule_weight == 0:
                break
            self.subrules.append(subrule)
            ptr += 4 + subrule.rule_size



import logging
from colorama import Fore, Style, init
from tqdm import tqdm


class TqdmLoggingHandler(logging.StreamHandler):
    def __init__(self):
        super().__init__()

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)

class ColorFormatter(logging.Formatter):
    # 初期化
    def __init__(self, fmt, datefmt):
        super().__init__(fmt, datefmt)
        self.FORMATS = {
            logging.DEBUG: self._color_format(Fore.BLUE, fmt),
            logging.INFO: self._color_format(Fore.GREEN, fmt),
            logging.WARNING: self._color_format(Fore.YELLOW, fmt),
            logging.ERROR: self._color_format(Fore.RED, fmt),
            logging.CRITICAL: self._color_format(Fore.RED + Style.BRIGHT, fmt),
        }
        self.datefmt = datefmt

    def _color_format(self, color, fmt):
        return Fore.LIGHTWHITE_EX + '[%(asctime)s]' + Style.RESET_ALL + f' {color}[%(levelname)s]' + Style.RESET_ALL + ' %(message)s'

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, self.datefmt)
        return formatter.format(record)

def setup_logger(name, debug=False):
    if debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    
    init()  # colorama
    fmt = '[%(asctime)s] [%(levelname)s] %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'

    logger = logging.getLogger(name)
    logger.setLevel(level)
    

    console_handler = TqdmLoggingHandler()
    console_handler.setFormatter(ColorFormatter(fmt, datefmt))

    # formatter = logging.Formatter(
    #     fmt='[%(asctime)s] [%(levelname)s] %(message)s',
    #     datefmt='%Y-%m-%d %H:%M:%S'
    # )
    # console_handler.setFormatter(formatter)

    if not logger.hasHandlers():
        logger.addHandler(console_handler)
    
    return logger

def suppress_logging(name):
    logging.getLogger(name).setLevel(logging.CRITICAL)


from typing import Tuple,List
import re
from defender2yara.defender.signature.base import BaseSig
from defender2yara.defender.signature.threat import ThreatBeginSig
from defender2yara.defender.signature.hstr import HSTR_EXT_SIGS
from defender2yara.util.utils import is_printable_ascii

MS_CARO_MALWARE_NAMING_PAT = r'^([A-Za-z0-0]+):([A-Za-z0-9_]+)(/([A-Za-z0-9_-]+))?((([\.!][A-Z0-9_]+)|([\.@!][A-Za-z0-9-_#]*))*)?$'

SEVERITY_MAP = {
    1: "Low",
    2: "Mid",
    4: "High",
    5: "Critical",
}

class Threat:
    """
    Manage threat defined by multiple signatures.
    """
    def __init__(self,sig:ThreatBeginSig):
        self.threat_begin_sig = sig
        self.threat_name = sig.threat_name
        self.threat_id = sig.threat_id
        self.category_id = sig.category_id
        
        if sig.severity_id in SEVERITY_MAP.keys():
            self.severity = SEVERITY_MAP[sig.severity_id]
        else:
            self.severity = sig.severity_id

        self.signatures = []
        self.hstr_signatures = []
        self.threat_type,\
        self.threat_platform,\
        self.threat_family,\
        self.threat_variant,\
        self.threat_suffixes = self.parse_threat_name(self.threat_name)

    def add_signature(self,sig:BaseSig):
        self.signatures.append(sig)
        if sig.sig_type == "SIGNATURE_TYPE_PEHSTR" or sig.sig_type in HSTR_EXT_SIGS:
            self.hstr_signatures.append(sig)

    @staticmethod
    def parse_threat_name(threat_name:str) -> Tuple[str,str,str,str,List[str]]:
        threat_type = ""
        threat_platform = ""
        threat_family = ""
        threat_variant = ""
        threat_suffixes = []
        
        if is_printable_ascii(threat_name):
            m = re.findall(MS_CARO_MALWARE_NAMING_PAT,threat_name)

            if not m:
                # print(threat_name)
                return \
                    threat_type,\
                    threat_platform,\
                    threat_family,\
                    threat_variant,\
                    threat_suffixes

            m = m[0]
            threat_type = m[0]
            threat_platform = m[1]
            threat_family = m[3]
            tail = m[4]
            if tail:
                tail_items = re.split(r'[\.@!]+',tail)
                for tail_item in tail_items:
                    if re.match(r'[A-Z0-9_]+',tail_item):
                        if tail_item == 'MSR' or re.match(r'MTB?',tail_item):
                            threat_suffixes.append(tail_item)
                        elif not threat_variant:
                            threat_variant = tail_item
                        else:
                            threat_suffixes.append(tail_item)
                    elif re.match(r'[A-Za-z0-9-_#]+',tail_item):
                        threat_suffixes.append(tail_item)
        
        return \
            threat_type,\
            threat_platform,\
            threat_family,\
            threat_variant,\
            threat_suffixes





from defender2yara.defender.signature.base import BaseSig
from defender2yara.defender.signature.hstr import HStrSig, HStrExtSig, HSTR_EXT_SIGS
from defender2yara.defender.signature.threat import ThreatBeginSig, ThreatEndSig


# ref: https://learn.microsoft.com/en-us/defender-xdr/malware-naming?view=o365-worldwide
from collections import defaultdict

SUFFIX = defaultdict(lambda: "an internal category used to refer to some threats")

SUFFIX['dam'] = "damaged malware"
SUFFIX['dll'] = "Dynamic Link Library component of a malware"
SUFFIX['dr'] = "dropper component of a malware"
SUFFIX['gen'] = "malware that is detected using a generic signature"
SUFFIX['kit'] = "virus constructor"
SUFFIX['ldr'] = "loader component of a malware"
SUFFIX['pak'] = "compressed malware"
SUFFIX['plugin'] = "plug-in component"
SUFFIX['remnants'] = "remnants of a virus"
SUFFIX['worm'] = "worm component of that malware"
SUFFIX['bit'] = "an internal category used to refer to some threats"
SUFFIX['cl'] = "an internal category used to refer to some threats"
SUFFIX['dha'] = "an internal category used to refer to some threats"
SUFFIX['pfn'] = "an internal category used to refer to some threats"
SUFFIX['plock'] = "an internal category used to refer to some threats"
SUFFIX['rfn'] = "an internal category used to refer to some threats"
SUFFIX['rootkit'] = "rootkit component of that malware"
SUFFIX['m'] = "worm mailers"
SUFFIX['mm'] = "mass mailer worm"
SUFFIX['MTB'] = "Microsoft Threat Behavior"
SUFFIX['MTB1'] = "Microsoft Threat Behavior"
SUFFIX['MTB6'] = "Microsoft Threat Behavior"
SUFFIX['MTB7'] = "Microsoft Threat Behavior"
SUFFIX['MTB8'] = "Microsoft Threat Behavior"
SUFFIX['MSR'] = "Microsoft Security Response"


import struct
from defender2yara.defender.constant import SIG_TYPES

class BaseSig:
    def __init__(self,data:bytes,ptr:int=0):
        self.parse_common_header(data,ptr)

    def parse_common_header(self,data:bytes,ptr:int=0):
        offset = 4
        self.sig_type_id = data[ptr+0]
        self.sig_type = SIG_TYPES[data[ptr+0]]
        self.size_low = data[ptr+1]
        self.size_high = struct.unpack('<H',data[ptr+2:ptr+4])[0]
        self.size = self.size_low | self.size_high << 8
        if self.size == 0xffffff:
            self.size = struct.unpack('<I',data[ptr+4:ptr+8])[0]
            offset += 4
        self.sig_data = data[ptr+offset:ptr+offset+self.size]


from collections import defaultdict

SIG_TYPES = defaultdict(lambda: "SIGNATURE_TYPE_UNKNOWN")
SIG_TYPES[0x01]= "SIGNATURE_TYPE_RESERVED"
SIG_TYPES[0x02]= "SIGNATURE_TYPE_VOLATILE_THREAT_INFO"
SIG_TYPES[0x03]= "SIGNATURE_TYPE_VOLATILE_THREAT_ID"
SIG_TYPES[0x11]= "SIGNATURE_TYPE_CKOLDREC"
SIG_TYPES[0x20]= "SIGNATURE_TYPE_KVIR32"
SIG_TYPES[0x21]= "SIGNATURE_TYPE_POLYVIR32"
SIG_TYPES[0x27]= "SIGNATURE_TYPE_NSCRIPT_NORMAL"
SIG_TYPES[0x28]= "SIGNATURE_TYPE_NSCRIPT_SP"
SIG_TYPES[0x29]= "SIGNATURE_TYPE_NSCRIPT_BRUTE"
SIG_TYPES[0x2C]= "SIGNATURE_TYPE_NSCRIPT_CURE"
SIG_TYPES[0x30]= "SIGNATURE_TYPE_TITANFLT"
SIG_TYPES[0x3D]= "SIGNATURE_TYPE_PEFILE_CURE"
SIG_TYPES[0x3E]= "SIGNATURE_TYPE_MAC_CURE"
SIG_TYPES[0x40]= "SIGNATURE_TYPE_SIGTREE"
SIG_TYPES[0x41]= "SIGNATURE_TYPE_SIGTREE_EXT"
SIG_TYPES[0x42]= "SIGNATURE_TYPE_MACRO_PCODE"
SIG_TYPES[0x43]= "SIGNATURE_TYPE_MACRO_SOURCE"
SIG_TYPES[0x44]= "SIGNATURE_TYPE_BOOT"
SIG_TYPES[0x49]= "SIGNATURE_TYPE_CLEANSCRIPT"
SIG_TYPES[0x4A]= "SIGNATURE_TYPE_TARGET_SCRIPT"
SIG_TYPES[0x50]= "SIGNATURE_TYPE_CKSIMPLEREC"
SIG_TYPES[0x51]= "SIGNATURE_TYPE_PATTMATCH"
SIG_TYPES[0x53]= "SIGNATURE_TYPE_RPFROUTINE"
SIG_TYPES[0x55]= "SIGNATURE_TYPE_NID"
SIG_TYPES[0x56]= "SIGNATURE_TYPE_GENSFX"
SIG_TYPES[0x57]= "SIGNATURE_TYPE_UNPLIB"
SIG_TYPES[0x58]= "SIGNATURE_TYPE_DEFAULTS"
SIG_TYPES[0x5B]= "SIGNATURE_TYPE_DBVAR"
SIG_TYPES[0x5C]= "SIGNATURE_TYPE_THREAT_BEGIN"
SIG_TYPES[0x5D]= "SIGNATURE_TYPE_THREAT_END"
SIG_TYPES[0x5E]= "SIGNATURE_TYPE_FILENAME"
SIG_TYPES[0x5F]= "SIGNATURE_TYPE_FILEPATH"
SIG_TYPES[0x60]= "SIGNATURE_TYPE_FOLDERNAME"
SIG_TYPES[0x61]= "SIGNATURE_TYPE_PEHSTR"
SIG_TYPES[0x62]= "SIGNATURE_TYPE_LOCALHASH"
SIG_TYPES[0x63]= "SIGNATURE_TYPE_REGKEY"
SIG_TYPES[0x64]= "SIGNATURE_TYPE_HOSTSENTRY"
SIG_TYPES[0x67]= "SIGNATURE_TYPE_STATIC"
SIG_TYPES[0x69]= "SIGNATURE_TYPE_LATENT_THREAT"
SIG_TYPES[0x6A]= "SIGNATURE_TYPE_REMOVAL_POLICY"
SIG_TYPES[0x6B]= "SIGNATURE_TYPE_WVT_EXCEPTION"
SIG_TYPES[0x6C]= "SIGNATURE_TYPE_REVOKED_CERTIFICATE"
SIG_TYPES[0x70]= "SIGNATURE_TYPE_TRUSTED_PUBLISHER"
SIG_TYPES[0x71]= "SIGNATURE_TYPE_ASEP_FILEPATH"
SIG_TYPES[0x73]= "SIGNATURE_TYPE_DELTA_BLOB"
SIG_TYPES[0x74]= "SIGNATURE_TYPE_DELTA_BLOB_RECINFO"
SIG_TYPES[0x75]= "SIGNATURE_TYPE_ASEP_FOLDERNAME"
SIG_TYPES[0x77]= "SIGNATURE_TYPE_PATTMATCH_V2"
SIG_TYPES[0x78]= "SIGNATURE_TYPE_PEHSTR_EXT"
SIG_TYPES[0x79]= "SIGNATURE_TYPE_VDLL_X86"
SIG_TYPES[0x7A]= "SIGNATURE_TYPE_VERSIONCHECK"
SIG_TYPES[0x7B]= "SIGNATURE_TYPE_SAMPLE_REQUEST"
SIG_TYPES[0x7C]= "SIGNATURE_TYPE_VDLL_X64"
SIG_TYPES[0x7E]= "SIGNATURE_TYPE_SNID"
SIG_TYPES[0x7F]= "SIGNATURE_TYPE_FOP"
SIG_TYPES[0x80]= "SIGNATURE_TYPE_KCRCE"
SIG_TYPES[0x83]= "SIGNATURE_TYPE_VFILE"
SIG_TYPES[0x84]= "SIGNATURE_TYPE_SIGFLAGS"
SIG_TYPES[0x85]= "SIGNATURE_TYPE_PEHSTR_EXT2"
SIG_TYPES[0x86]= "SIGNATURE_TYPE_PEMAIN_LOCATOR"
SIG_TYPES[0x87]= "SIGNATURE_TYPE_PESTATIC"
SIG_TYPES[0x88]= "SIGNATURE_TYPE_UFSP_DISABLE"
SIG_TYPES[0x89]= "SIGNATURE_TYPE_FOPEX"
SIG_TYPES[0x8A]= "SIGNATURE_TYPE_PEPCODE"
SIG_TYPES[0x8B]= "SIGNATURE_TYPE_IL_PATTERN"
SIG_TYPES[0x8C]= "SIGNATURE_TYPE_ELFHSTR_EXT"
SIG_TYPES[0x8D]= "SIGNATURE_TYPE_MACHOHSTR_EXT"
SIG_TYPES[0x8E]= "SIGNATURE_TYPE_DOSHSTR_EXT"
SIG_TYPES[0x8F]= "SIGNATURE_TYPE_MACROHSTR_EXT"
SIG_TYPES[0x90]= "SIGNATURE_TYPE_TARGET_SCRIPT_PCODE"
SIG_TYPES[0x91]= "SIGNATURE_TYPE_VDLL_IA64"
SIG_TYPES[0x95]= "SIGNATURE_TYPE_PEBMPAT"
SIG_TYPES[0x96]= "SIGNATURE_TYPE_AAGGREGATOR"
SIG_TYPES[0x97]= "SIGNATURE_TYPE_SAMPLE_REQUEST_BY_NAME"
SIG_TYPES[0x98]= "SIGNATURE_TYPE_REMOVAL_POLICY_BY_NAME"
SIG_TYPES[0x99]= "SIGNATURE_TYPE_TUNNEL_X86"
SIG_TYPES[0x9A]= "SIGNATURE_TYPE_TUNNEL_X64"
SIG_TYPES[0x9B]= "SIGNATURE_TYPE_TUNNEL_IA64"
SIG_TYPES[0x9C]= "SIGNATURE_TYPE_VDLL_ARM"
SIG_TYPES[0x9D]= "SIGNATURE_TYPE_THREAD_X86"
SIG_TYPES[0x9E]= "SIGNATURE_TYPE_THREAD_X64"
SIG_TYPES[0x9F]= "SIGNATURE_TYPE_THREAD_IA64"
SIG_TYPES[0xA0]= "SIGNATURE_TYPE_FRIENDLYFILE_SHA256"
SIG_TYPES[0xA1]= "SIGNATURE_TYPE_FRIENDLYFILE_SHA512"
SIG_TYPES[0xA2]= "SIGNATURE_TYPE_SHARED_THREAT"
SIG_TYPES[0xA3]= "SIGNATURE_TYPE_VDM_METADATA"
SIG_TYPES[0xA4]= "SIGNATURE_TYPE_VSTORE"
SIG_TYPES[0xA5]= "SIGNATURE_TYPE_VDLL_SYMINFO"
SIG_TYPES[0xA6]= "SIGNATURE_TYPE_IL2_PATTERN"
SIG_TYPES[0xA7]= "SIGNATURE_TYPE_BM_STATIC"
SIG_TYPES[0xA8]= "SIGNATURE_TYPE_BM_INFO"
SIG_TYPES[0xA9]= "SIGNATURE_TYPE_NDAT"
SIG_TYPES[0xAA]= "SIGNATURE_TYPE_FASTPATH_DATA"
SIG_TYPES[0xAB]= "SIGNATURE_TYPE_FASTPATH_SDN"
SIG_TYPES[0xAC]= "SIGNATURE_TYPE_DATABASE_CERT"
SIG_TYPES[0xAD]= "SIGNATURE_TYPE_SOURCE_INFO"
SIG_TYPES[0xAE]= "SIGNATURE_TYPE_HIDDEN_FILE"
SIG_TYPES[0xAF]= "SIGNATURE_TYPE_COMMON_CODE"
SIG_TYPES[0xB0]= "SIGNATURE_TYPE_VREG"
SIG_TYPES[0xB1]= "SIGNATURE_TYPE_NISBLOB"
SIG_TYPES[0xB2]= "SIGNATURE_TYPE_VFILEEX"
SIG_TYPES[0xB3]= "SIGNATURE_TYPE_SIGTREE_BM"
SIG_TYPES[0xB4]= "SIGNATURE_TYPE_VBFOP"
SIG_TYPES[0xB5]= "SIGNATURE_TYPE_VDLL_META"
SIG_TYPES[0xB6]= "SIGNATURE_TYPE_TUNNEL_ARM"
SIG_TYPES[0xB7]= "SIGNATURE_TYPE_THREAD_ARM"
SIG_TYPES[0xB8]= "SIGNATURE_TYPE_PCODEVALIDATOR"
SIG_TYPES[0xBA]= "SIGNATURE_TYPE_MSILFOP"
SIG_TYPES[0xBB]= "SIGNATURE_TYPE_KPAT"
SIG_TYPES[0xBC]= "SIGNATURE_TYPE_KPATEX"
SIG_TYPES[0xBD]= "SIGNATURE_TYPE_LUASTANDALONE"
SIG_TYPES[0xBE]= "SIGNATURE_TYPE_DEXHSTR_EXT"
SIG_TYPES[0xBF]= "SIGNATURE_TYPE_JAVAHSTR_EXT"
SIG_TYPES[0xC0]= "SIGNATURE_TYPE_MAGICCODE"
SIG_TYPES[0xC1]= "SIGNATURE_TYPE_CLEANSTORE_RULE"
SIG_TYPES[0xC2]= "SIGNATURE_TYPE_VDLL_CHECKSUM"
SIG_TYPES[0xC3]= "SIGNATURE_TYPE_THREAT_UPDATE_STATUS"
SIG_TYPES[0xC4]= "SIGNATURE_TYPE_VDLL_MSIL"
SIG_TYPES[0xC5]= "SIGNATURE_TYPE_ARHSTR_EXT"
SIG_TYPES[0xC6]= "SIGNATURE_TYPE_MSILFOPEX"
SIG_TYPES[0xC7]= "SIGNATURE_TYPE_VBFOPEX"
SIG_TYPES[0xC8]= "SIGNATURE_TYPE_FOP64"
SIG_TYPES[0xC9]= "SIGNATURE_TYPE_FOPEX64"
SIG_TYPES[0xCA]= "SIGNATURE_TYPE_JSINIT"
SIG_TYPES[0xCB]= "SIGNATURE_TYPE_PESTATICEX"
SIG_TYPES[0xCC]= "SIGNATURE_TYPE_KCRCEX"
SIG_TYPES[0xCD]= "SIGNATURE_TYPE_FTRIE_POS"
SIG_TYPES[0xCE]= "SIGNATURE_TYPE_NID64"
SIG_TYPES[0xCF]= "SIGNATURE_TYPE_MACRO_PCODE64"
SIG_TYPES[0xD0]= "SIGNATURE_TYPE_BRUTE"
SIG_TYPES[0xD1]= "SIGNATURE_TYPE_SWFHSTR_EXT"
SIG_TYPES[0xD2]= "SIGNATURE_TYPE_REWSIGS"
SIG_TYPES[0xD3]= "SIGNATURE_TYPE_AUTOITHSTR_EXT"
SIG_TYPES[0xD4]= "SIGNATURE_TYPE_INNOHSTR_EXT"
SIG_TYPES[0xD5]= "SIGNATURE_TYPE_ROOTCERTSTORE"
SIG_TYPES[0xD6]= "SIGNATURE_TYPE_EXPLICITRESOURCE"
SIG_TYPES[0xD7]= "SIGNATURE_TYPE_CMDHSTR_EXT"
SIG_TYPES[0xD8]= "SIGNATURE_TYPE_FASTPATH_TDN"
SIG_TYPES[0xD9]= "SIGNATURE_TYPE_EXPLICITRESOURCEHASH"
SIG_TYPES[0xDA]= "SIGNATURE_TYPE_FASTPATH_SDN_EX"
SIG_TYPES[0xDB]= "SIGNATURE_TYPE_BLOOM_FILTER"
SIG_TYPES[0xDC]= "SIGNATURE_TYPE_RESEARCH_TAG"
SIG_TYPES[0xDE]= "SIGNATURE_TYPE_ENVELOPE"
SIG_TYPES[0xDF]= "SIGNATURE_TYPE_REMOVAL_POLICY64"
SIG_TYPES[0xE0]= "SIGNATURE_TYPE_REMOVAL_POLICY64_BY_NAME"
SIG_TYPES[0xE1]= "SIGNATURE_TYPE_VDLL_META_X64"
SIG_TYPES[0xE2]= "SIGNATURE_TYPE_VDLL_META_ARM"
SIG_TYPES[0xE3]= "SIGNATURE_TYPE_VDLL_META_MSIL"
SIG_TYPES[0xE4]= "SIGNATURE_TYPE_MDBHSTR_EXT"
SIG_TYPES[0xE5]= "SIGNATURE_TYPE_SNIDEX"
SIG_TYPES[0xE6]= "SIGNATURE_TYPE_SNIDEX2"
SIG_TYPES[0xE7]= "SIGNATURE_TYPE_AAGGREGATOREX"
SIG_TYPES[0xE8]= "SIGNATURE_TYPE_PUA_APPMAP"
SIG_TYPES[0xE9]= "SIGNATURE_TYPE_PROPERTY_BAG"
SIG_TYPES[0xEA]= "SIGNATURE_TYPE_DMGHSTR_EXT"
SIG_TYPES[0xEB]= "SIGNATURE_TYPE_DATABASE_CATALOG"


import sys
from .main import main
import toml
from pathlib import Path

from defender2yara.util.logging import setup_logger, suppress_logging
from defender2yara.util.utils import is_validate_filesize


def get_version():
    pyproject_path = Path(__file__).parent.parent / 'pyproject.toml'
    pyproject_data = toml.load(pyproject_path)
    return pyproject_data['tool']['poetry']['version']


def run():
    import argparse
    parser = argparse.ArgumentParser(
        description="Convert Microsoft Defender Antivirus Signatures(VDM) to YARA rules.",
        usage="defender2yara [options]")

    parser.add_argument('-v', '--version', action='store_true', help="show defender2yara version")
    parser.add_argument('-l','--latest_signature_version', action='store_true', default=False, help="show latest signature version")
    parser.add_argument('-o','--output', default='./rules',help="output directory for YARA rules [default: ./rules]" )
    parser.add_argument('-d','--download',action='store_true' ,required=False,default=False,help="only download the latest signature database")
    parser.add_argument('-c','--cache', default='./cache',help="directory to save signature database(vdm/dll files) [default: ./cache]" )
    parser.add_argument('-s','--single_file',action='store_true',default=False,help="export YARA rules into a single file")
    parser.add_argument('--header_check',action='store_true',default=False,help="add file header check to generated YARA rules")
    parser.add_argument('--filesize_check',required=False,type=str,default="20MB",help="add filesize check to generated YARA rules")
    parser.add_argument('--no_filesize_check',action='store_true',required=False,default=False,help="remove filesize check from generated YARA rules")
    parser.add_argument('--mpam',required=False,type=str,help="manually specify the path of mpam-fe.exe")
    parser.add_argument('--base',required=False,type=str,help="manually specify the path of mpa{v|s}base.vdm")
    parser.add_argument('--delta',required=False,type=str,help="manually specify the path of mpa{v|s}dlta.vdm")
    parser.add_argument('--proxy',help="use a proxy to download signatures (e.g. http://localhost:8000)")
    parser.add_argument('--debug', action='store_true', default=False, help="print detailed logs")
    parser.add_argument('--suppress', action='store_true', default=False, help="suppress all logs")
    
    args = parser.parse_args()
    
    if args.version:
        print(f"version: {get_version()}")
        sys.exit(0)

    if args.base and args.download:
        sys.stderr.write("[!] --download option and --base option can not use together.")
        parser.print_help()
        sys.exit(1)
    
    if args.download and args.mpam:
        sys.stderr.write("[!] --download option and --mpam option can not use together.")
        parser.print_help()
        sys.exit(1)
    
    if args.mpam and (args.base or args.delta):
        sys.stderr.write("[!] --mpam option and --base or --delta option can not use together.")
        parser.print_help()
        sys.exit(1)
    
    if args.delta and not args.base:
        sys.stderr.write("[!] --delta option requires --base option.")
        parser.print_help()
        sys.exit()

    if args.suppress and args.debug:
        sys.stderr.write("[!] --suppress option and --debug option can not use together.")
        parser.print_help()
        sys.exit(1)

    if not is_validate_filesize(args.filesize_check):
        sys.stderr.write("[!] Invalid filesize format. Use integer with postfix KB or MB.")
        parser.print_help()
        sys.exit(1)

    if args.no_filesize_check:
        # remove filesize check from generated yara rules.
        args.filesize_check = ""

    setup_logger(__package__, args.debug)

    if args.suppress:
        suppress_logging(__package__)

    main(args)


if __name__ == "__main__":
    run()